/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"

/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse) {
    std::shared_ptr<Query> query = std::make_shared<Query>();
    if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse)) {
        // check table
        query->tables = std::move(x->tabs);
        for (auto& table : query->tables) {
            if (!sm_manager_->db_.is_table(table)) {
                throw TableNotFoundError(table);
            }
        }

        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);

        // check select fields
        bool have_aggregate = false;
        for (auto& sv_sel_col : x->cols) {
            TabCol sel_col = {.tab_name = sv_sel_col->tab_name,
                              .col_name = sv_sel_col->col_name,
                              .alias_name = sv_sel_col->alias_name};
            sel_col.aggr_op = convert_sv_aggr_op(sv_sel_col->aggr_op);
            have_aggregate |= sel_col.aggr_op != AGGR_NONE;

            if (sel_col.col_name == "*") {
                // FIXME: Not a good way to handle this
                if (!sel_col.alias_name.empty() && sel_col.aggr_op == AggrOp::AGGR_NONE) {
                    throw InvalidAliasNameError();
                }

                if (!sel_col.tab_name.empty()) {
                    bool table_found = false;
                    for (auto& table : query->tables) {
                        if (table == sel_col.tab_name) {
                            table_found = true;
                            break;
                        }
                    }
                    if (!table_found) {
                        throw TableNotFoundError(sel_col.tab_name);
                    }
                }
                switch (sel_col.aggr_op) {
                    case AggrOp::AGGR_NONE: {
                        // SELECT STAR
                        for (auto& col : all_cols) {
                            if (sel_col.tab_name.empty() || sel_col.tab_name == col.tab_name) {
                                TabCol temp_col = {.tab_name = col.tab_name, .col_name = col.name};
                                query->cols.push_back(temp_col);
                            }
                        }
                        break;
                    }
                    case AggrOp::AGGR_COUNT: {
                        // SELECT COUNT(*)
                        query->cols.push_back(sel_col);
                        break;
                    }
                    default: {
                        throw InvalidAggrOpError();
                    }
                }
            } else {
                sel_col = check_column(all_cols, sel_col);  // 列元数据校验
                query->cols.push_back(sel_col);
            }
        }

        // 处理where条件
        get_where_clause(x->where_conds, query->where_conds);
        check_where_clause(query->tables, query->where_conds);

        // groupby
        get_groupby_clause(x->groupby_cols, query->groupby_cols, all_cols);
        // having
        get_having_clause(x->having_conds, query->having_conds);
        check_having_clause(query->tables, query->having_conds);
        check_groupby_clause(query->groupby_cols, have_aggregate, query->cols, query->having_conds);

        // 处理order列
        if (x->order != nullptr) {
            for (auto& sv_order_col : x->order->cols) {
                TabCol order_col = {.tab_name = sv_order_col->col->tab_name, .col_name = sv_order_col->col->col_name};
                query->order_cols.push_back(order_col);
            }
        }
        for (auto& order_col : query->order_cols) {
            order_col = check_column(all_cols, order_col);
        }
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
        if (!sm_manager_->db_.is_table(x->tab_name)) {
            throw TableNotFoundError(x->tab_name);
        }

        get_update_clause(x->tab_name, x->set_clauses, query->set_clauses);
        check_update_clause(x->tab_name, query->set_clauses);

        get_where_clause(x->conds, query->where_conds);
        check_where_clause({x->tab_name}, query->where_conds);
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
        if (!sm_manager_->db_.is_table(x->tab_name)) {
            throw TableNotFoundError(x->tab_name);
        }

        get_where_clause(x->conds, query->where_conds);
        check_where_clause({x->tab_name}, query->where_conds);
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
        if (!sm_manager_->db_.is_table(x->tab_name)) {
            throw TableNotFoundError(x->tab_name);
        }

        // 处理insert 的values值
        for (auto& sv_val : x->vals) {
            query->values.push_back(convert_sv_value(sv_val));
        }
    } else {
        // do nothing
    }
    query->parse = std::move(parse);
    return query;
}

TabCol Analyze::check_column(const std::vector<ColMeta>& all_cols, TabCol target) {
    if (target.tab_name.empty()) {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto& col : all_cols) {
            if (col.name == target.col_name) {
                if (!tab_name.empty()) {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty() && !(target.aggr_op == AggrOp::AGGR_COUNT && target.col_name == "*")) {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    } else {
        bool exist = false;
        for (auto& col : all_cols) {
            if (col.tab_name != target.tab_name) {
                continue;
            }
            if (col.name == target.col_name) {
                exist = true;
                break;
            }
        }
        if (!exist) {
            throw ColumnNotFoundError(target.col_name);
        }
    }
    return target;
}

void Analyze::get_all_cols(const std::vector<std::string>& tab_names, std::vector<ColMeta>& all_cols) {
    for (auto& sel_tab_name : tab_names) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto& sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
}

void Analyze::get_where_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>>& sv_conds,
                               std::vector<Condition>& conds) {
    conds.clear();
    for (auto& expr : sv_conds) {
        if (expr->lhs->aggr_op != ast::SV_AGGR_NONE) {
            throw InvalidAggrOpError();
        }

        Condition cond;
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name};
        cond.lhs_col.aggr_op = AggrOp::AGGR_NONE;
        cond.op = convert_sv_comp_op(expr->op);
        if (!expr->rhs_vals.empty()) {
            if (cond.op != CompOp::OP_IN && expr->rhs_vals.size() > 1) {
                throw InternalError("Unexpected op in condition");
            }
            cond.rhs_type = RhsType::LIST;
            cond.rhs_val_list = std::vector<Value>();
            for (auto& sv_val : expr->rhs_vals) {
                cond.rhs_val_list.emplace_back(convert_sv_value(sv_val));
            }
        } else if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.rhs_type = RhsType::VALUE;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            if (rhs_col->aggr_op != ast::SV_AGGR_NONE) {
                throw InvalidAggrOpError();
            }
            cond.rhs_type = RhsType::COLUMN;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
            cond.rhs_col.aggr_op = AggrOp::AGGR_NONE;
        } else if (auto rhs_sub_query = std::dynamic_pointer_cast<ast::SelectStmt>(expr->rhs)) {
            cond.rhs_type = RhsType::SUB_QUERY;
            cond.rhs_sub_query = do_analyze(rhs_sub_query);
        } else {
            throw InternalError("Unexpected expr type");
        }
        conds.push_back(cond);
    }
}

void Analyze::check_where_clause(const std::vector<std::string>& tab_names, std::vector<Condition>& conds) {
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto& cond : conds) {
        // Infer table name from column name
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if (cond.rhs_type == RhsType::COLUMN) {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }

        ColType lhs_type;
        ColType rhs_type;

        TabMeta& lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        lhs_type = lhs_col->type;
        if (cond.rhs_type == RhsType::VALUE) {
            rhs_type = cond.rhs_val.type;
        } else if (cond.rhs_type == RhsType::COLUMN) {
            TabMeta& rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        } else if (cond.rhs_type == RhsType::LIST) {
            rhs_type = cond.rhs_val_list[0].type;
            for (auto& val : cond.rhs_val_list) {
                if (val.type != rhs_type && !(val.type != ColType::TYPE_STRING && rhs_type != ColType::TYPE_STRING)) {
                    throw IncompatibleTypeError(coltype2str(rhs_type), coltype2str(val.type));
                }
            }
        } else {
            auto size = cond.rhs_sub_query->cols.size();
            if (size != 1) {
                throw InvalidFieldsInSubQueryError();
            }
            // skip type check for sub-query
            rhs_type = lhs_type;
        }
        if (lhs_type != rhs_type && !(lhs_type != ColType::TYPE_STRING && rhs_type != ColType::TYPE_STRING)) {
            throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
        }
        if (cond.rhs_type == RhsType::VALUE) {
            cond.rhs_val.init_raw(lhs_col->len);
        }
    }
}

void Analyze::get_groupby_clause(const std::vector<std::shared_ptr<ast::Col>>& sv_groupby_cols,
                                 std::vector<TabCol>& groupby_cols, const std::vector<ColMeta>& all_cols) {
    groupby_cols.clear();
    for (auto& col : sv_groupby_cols) {
        if (col->aggr_op != ast::SV_AGGR_NONE) {
            throw InvalidAggrOpError();
        }
        TabCol groupby_col = {.tab_name = col->tab_name, .col_name = col->col_name};
        groupby_col.aggr_op = AggrOp::AGGR_NONE;
        groupby_col = check_column(all_cols, groupby_col);
        groupby_cols.push_back(groupby_col);
    }
}

void Analyze::get_having_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>>& sv_having_conds,
                                std::vector<Condition>& having_conds) {
    having_conds.clear();
    for (auto& expr : sv_having_conds) {
        Condition cond;
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name};
        cond.lhs_col.aggr_op = convert_sv_aggr_op(expr->lhs->aggr_op);
        cond.op = convert_sv_comp_op(expr->op);
        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.rhs_type = RhsType::VALUE;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            cond.rhs_type = RhsType::COLUMN;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
            cond.rhs_col.aggr_op = convert_sv_aggr_op(rhs_col->aggr_op);
        } else {
            throw InternalError("Unexpected expr type");
        }
        having_conds.push_back(cond);
    }
}

void Analyze::check_having_clause(const std::vector<std::string>& tab_name, std::vector<Condition>& having_conds) {
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_name, all_cols);

    for (auto& cond : having_conds) {
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        if (cond.rhs_type == RhsType::COLUMN) {
            cond.rhs_col = check_column(all_cols, cond.rhs_col);
        }

        ColType lhs_type;
        ColType rhs_type;
        size_t len = 0;

        if (cond.lhs_col.col_name == "*") {
            if (cond.lhs_col.aggr_op != AggrOp::AGGR_COUNT) {
                throw InvalidAggrOpError();
            }
            lhs_type = ColType::TYPE_INT;
            len = sizeof(int);
        } else {
            TabMeta& lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
            auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
            lhs_type = lhs_col->type;
            len = lhs_col->len;
        }
        if (cond.rhs_type == RhsType::VALUE) {
            rhs_type = cond.rhs_val.type;
        } else if (cond.rhs_type == RhsType::COLUMN) {
            if (cond.rhs_col.col_name == "*") {
                if (cond.lhs_col.aggr_op != AggrOp::AGGR_COUNT) {
                    throw InvalidAggrOpError();
                }
                rhs_type = ColType::TYPE_INT;
            } else {
                TabMeta& rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
                auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
                rhs_type = rhs_col->type;
            }
        } else {
            throw InternalError("Unexpected rhs type");
        }

        if (lhs_type != rhs_type && !(lhs_type != ColType::TYPE_STRING && rhs_type != ColType::TYPE_STRING)) {
            throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
        }
        if (cond.rhs_type == RhsType::VALUE) {
            cond.rhs_val.init_raw(len);
        }
    }
}

void Analyze::check_groupby_clause(const std::vector<TabCol>& groupby_cols, bool have_aggregate,
                                   const std::vector<TabCol>& sel_cols, const std::vector<Condition>& having_conds) {
    if (groupby_cols.empty() && !having_conds.empty()) {
        throw HavingWithoutGroupByError();
    }
    if (groupby_cols.empty()) {
        return;
    }

    for (auto& sel_col : sel_cols) {
        if (sel_col.aggr_op != AGGR_NONE) {
            continue;
        }
        bool found = false;
        for (auto& groupby_col : groupby_cols) {
            if (sel_col.tab_name == groupby_col.tab_name && sel_col.col_name == groupby_col.col_name) {
                found = true;
                break;
            }
        }
        if (!found) {
            throw ColumnNotInGroupByError(sel_col.col_name);
        }
    }

    for (auto& cond : having_conds) {
        bool lhs_found = false;
        bool rhs_found = cond.rhs_type == RhsType::VALUE;
        for (auto& groupby_col : groupby_cols) {
            if ((cond.lhs_col.tab_name == groupby_col.tab_name && cond.lhs_col.col_name == groupby_col.col_name) ||
                cond.lhs_col.aggr_op != AggrOp::AGGR_NONE) {
                lhs_found = true;
            }
            if (cond.rhs_type == RhsType::COLUMN &&
                ((cond.rhs_col.tab_name == groupby_col.tab_name && cond.rhs_col.col_name == groupby_col.col_name) ||
                 cond.rhs_col.aggr_op != AggrOp::AGGR_NONE)) {
                rhs_found = true;
            }
        }
        if (!lhs_found || !rhs_found) {
            throw ColumnNotInGroupByError(cond.lhs_col.col_name);
        }
    }
}

void Analyze::get_update_clause(const std::string& tab_name,
                                const std::vector<std::shared_ptr<ast::SetClause>>& sv_set_clauses,
                                std::vector<SetClause>& set_clauses) {
    set_clauses.clear();
    for (auto& sv_set_clause : sv_set_clauses) {
        SetClause set_clause;
        set_clause.lhs = {.tab_name = tab_name, .col_name = sv_set_clause->col_name};
        set_clause.rhs = convert_sv_value(sv_set_clause->val);
        if (sv_set_clause->rhs_has_lhs == true) {
            set_clause.rhs_col = {.tab_name = tab_name, .col_name = sv_set_clause->col_right};
            set_clause.rhs_has_lhs=true;
            set_clause.operation=sv_set_clause->operation;
        }
        set_clauses.push_back(set_clause);
    }
}

void Analyze::check_update_clause(const std::string& tab_name, std::vector<SetClause>& set_clauses) {
    std::vector<ColMeta> all_cols;
    get_all_cols({tab_name}, all_cols);
    for (auto& set_clause : set_clauses) {
        check_column(all_cols, set_clause.lhs);
        TabMeta& tab = sm_manager_->db_.get_table(tab_name);
        auto col = tab.get_col(set_clause.lhs.col_name);
        if (col->type != set_clause.rhs.type &&
            !(col->type != ColType::TYPE_STRING && set_clause.rhs.type != ColType::TYPE_STRING)) {
            throw IncompatibleTypeError(coltype2str(col->type), coltype2str(set_clause.rhs.type));
        }
        if (col->type == ColType::TYPE_INT && set_clause.rhs.type == ColType::TYPE_FLOAT) {
            set_clause.rhs.set_int((int)set_clause.rhs.float_val);
        }
        if (col->type == ColType::TYPE_FLOAT && set_clause.rhs.type == ColType::TYPE_INT) {
            set_clause.rhs.set_float((float)set_clause.rhs.int_val);
        }
        set_clause.rhs.init_raw(col->len);
    }
}

Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value>& sv_val) {
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
        val.set_int(int_lit->val);
    } else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
        val.set_float(float_lit->val);
    } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
        val.set_str(str_lit->val);
    } else {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
    std::map<ast::SvCompOp, CompOp> m = {{ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
                                         {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
                                         {ast::SV_OP_IN, OP_IN}};
    return m.at(op);
}

AggrOp Analyze::convert_sv_aggr_op(ast::SvAggrOp aggr_op) {
    std::map<ast::SvAggrOp, AggrOp> m = {
        {ast::SV_AGGR_NONE, AGGR_NONE}, {ast::SV_AGGR_COUNT, AGGR_COUNT}, {ast::SV_AGGR_SUM, AGGR_SUM},
        {ast::SV_AGGR_MAX, AGGR_MAX},   {ast::SV_AGGR_MIN, AGGR_MIN},
    };
    return m.at(aggr_op);
}