/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cstring>
#include <memory>
#include <string>
#include <vector>

#include "defs.h"
#include "record/rm_defs.h"

enum AggrOp {
    AGGR_NONE,
    AGGR_COUNT,
    AGGR_SUM,
    AGGR_MAX,
    AGGR_MIN,
};

class AggrOpHelper {
public:
    static std::string ToString(AggrOp op) {
        switch (op) {
            case AggrOp::AGGR_NONE:
                return "";
            case AggrOp::AGGR_COUNT:
                return "COUNT";
            case AggrOp::AGGR_SUM:
                return "SUM";
            case AggrOp::AGGR_MAX:
                return "MAX";
            case AggrOp::AGGR_MIN:
                return "MIN";
            default:
                throw std::runtime_error("Unknown AggrOp");
        }
    }
};
struct TabCol {
    std::string tab_name;
    std::string col_name;
    std::string alias_name;
    AggrOp aggr_op;

    friend bool operator<(const TabCol& x, const TabCol& y) {
        auto p1 = std::make_pair(x.tab_name, x.col_name);
        auto p2 = std::make_pair(y.tab_name, y.col_name);
        return p1 < p2 || (p1 == p2 && x.aggr_op < y.aggr_op);
    }

    friend bool operator==(const TabCol& x, const TabCol& y) {
        return std::make_pair(x.tab_name, x.col_name) == std::make_pair(y.tab_name, y.col_name) &&
            x.aggr_op == y.aggr_op;
    }
};

struct Value {
    ColType type;  // type of value
    union {
        int int_val;      // int value
        float float_val;  // float value
    };
    char* raw_ptr{(char*)&int_val};

    std::string str_val;            // string value
    std::shared_ptr<RmRecord> raw;  // raw record buffer

    Value() : type(ColType::TYPE_INT), int_val(0), raw(nullptr) {}
    Value(int int_val_) : type(ColType::TYPE_INT), int_val(int_val_), raw(nullptr) {}
    Value(float float_val_) : type(ColType::TYPE_FLOAT), float_val(float_val_), raw(nullptr) {}
    Value(ColType type_) : type(type_), raw(nullptr) {}

    void set_int(int int_val_) {
        type = TYPE_INT;
        int_val = int_val_;
    }

    void set_float(float float_val_) {
        type = TYPE_FLOAT;
        float_val = float_val_;
    }

    void set_str(std::string str_val_) {
        type = TYPE_STRING;
        str_val = std::move(str_val_);
    }

    void init_raw(int len) {
        // assert(raw == nullptr);
        raw = std::make_shared<RmRecord>(len);
        if (type == TYPE_INT) {
            assert(len == sizeof(int));
            *(int*)(raw->data) = int_val;
        } else if (type == TYPE_FLOAT) {
            assert(len == sizeof(float));
            *(float*)(raw->data) = float_val;
        } else if (type == TYPE_STRING) {
            if (len < (int)str_val.size()) {
                throw StringOverflowError();
            }
            memset(raw->data, 0, len);
            memcpy(raw->data, str_val.c_str(), str_val.size());
        }
    }

    inline void ConvertRecordToValue(char* data, ColType type, const size_t offset, const size_t len) {
        switch (type) {
            case ColType::TYPE_INT:
                set_int(*reinterpret_cast<int*>(data + offset));
                break;
            case ColType::TYPE_FLOAT:
                set_float(*reinterpret_cast<float*>(data + offset));
                break;
            case ColType::TYPE_STRING:
                set_str(std::string(data + offset, len));
                break;
            default:
                throw InternalError("unknown type");
        }
    }

    int Compare(const Value& rhs) { return *this < rhs ? -1 : (*this == rhs ? 0 : 1); }

    Value& operator=(const Value& rhs) {
        if (this == &rhs) {
            return *this;
        }
        type = rhs.type;
        switch (type) {
            case TYPE_INT:
                int_val = rhs.int_val;
                break;
            case TYPE_FLOAT:
                float_val = rhs.float_val;
                break;
            case TYPE_STRING:
                str_val = rhs.str_val;
                break;
            default:
                throw InternalError("unknown type");
        }
        return *this;
    }

    bool operator<(const Value& rhs) const {
        if (type != rhs.type && !(type != ColType::TYPE_STRING && rhs.type != ColType::TYPE_STRING)) {
            throw IncompatibleTypeError(coltype2str(type), coltype2str(rhs.type));
        }
        switch (type) {
            case TYPE_INT:
                return type == rhs.type ? int_val < rhs.int_val : int_val < rhs.float_val;
            case TYPE_FLOAT:
                return type == rhs.type ? float_val < rhs.float_val : float_val < rhs.int_val;
            case TYPE_STRING:
                return strncmp(str_val.c_str(), rhs.str_val.c_str(), std::max(str_val.size(), rhs.str_val.size())) < 0;
            default:
                throw InternalError("unknown type");
        }
    }

    bool operator==(const Value& rhs) const {
        if (type != rhs.type && !(type != ColType::TYPE_STRING && rhs.type != ColType::TYPE_STRING)) {
            throw IncompatibleTypeError(coltype2str(type), coltype2str(rhs.type));
        }
        switch (type) {
            case TYPE_INT:
                // return type == rhs.type ? int_val == rhs.int_val : (float)int_val == rhs.float_val;
                return type == rhs.type ? memcmp(raw_ptr, rhs.raw_ptr, sizeof(int)) == 0
                                        : (float)*(int*)raw_ptr == rhs.float_val;
            case TYPE_FLOAT:
                // return type == rhs.type ? float_val == rhs.float_val : float_val == (float)rhs.int_val;
                return type == rhs.type ? memcmp(raw_ptr, rhs.raw_ptr, sizeof(float)) == 0
                                        : *(float*)raw_ptr == (float)rhs.int_val;
            case TYPE_STRING:
                return strncmp(str_val.c_str(), rhs.str_val.c_str(), std::max(str_val.size(), rhs.str_val.size())) == 0;
            default:
                throw InternalError("unknown type");
        }
    }

    bool operator<=(const Value& rhs) const { return (*this < rhs) || (*this == rhs); }

    bool operator>(const Value& rhs) const { return !(*this < rhs) && !(*this == rhs); }

    bool operator>=(const Value& rhs) const { return !(*this < rhs); }

    bool operator!=(const Value& rhs) const { return !(*this == rhs); }
};

enum CompOp {
    OP_EQ,
    OP_NE,
    OP_LT,
    OP_GT,
    OP_LE,
    OP_GE,
    OP_IN,
};

enum RhsType {
    VALUE,
    COLUMN,
    LIST,
    SUB_QUERY,
};

class Query;

struct Condition {
    TabCol lhs_col;  // left-hand side column
    CompOp op;       // comparison operator
    RhsType rhs_type;
    TabCol rhs_col;  // right-hand side column

    Value rhs_val;  // right-hand side value
    std::vector<Value> rhs_val_list;
    std::shared_ptr<Query> rhs_sub_query;

    friend bool operator==(const Condition& x, const Condition& y) {
        if (x.lhs_col == y.lhs_col && x.op == y.op && x.rhs_type == y.rhs_type) {
            if (x.rhs_type == RhsType::VALUE) {
                if (x.rhs_val == y.rhs_val) {
                    return true;
                }
            }
            else if (x.rhs_type == RhsType::COLUMN) {
                if (x.rhs_col == y.rhs_col) {
                    return true;
                }
            }
        }
        return false;
    }
};

struct SetClause {
    TabCol lhs;
    TabCol rhs_col;
    Value rhs;
    bool rhs_has_lhs = false; //false:update ... set score=90 ... ; true:update ... set score=score+5 ...;
    int operation = 0; //0:equal; 1:add; 2:minus;
};