/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class DeleteExecutor : public AbstractExecutor {
public:
    DeleteExecutor(SmManager* sm_manager, const std::string& tab_name, std::vector<Condition> conds,
                   std::vector<Rid> rids, Context* context, Aggregate_cache* cache_) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;

        cache = cache_;
    }

    std::string getType() override { return "DeleteExecutor"; }

    std::unique_ptr<RmRecord> Next() override {
        std::vector<std::pair<IndexMeta, IxIndexHandle*>> index_map;
        for (auto& index : tab_.indexes) {
            index_map.emplace_back(
                index,
                sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get());
        }

        for (auto rid : rids_) {
            auto record = fh_->get_record(rid, context_);
            auto data = record->data;
            fh_->delete_record(rid, context_);

            // TODO: 这里可以避免频繁的内存分配
            for (auto& [index, ih] : index_map) {
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for (size_t i = 0; i < index.col_num; ++i) {
                    memcpy(key + offset, data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->delete_entry(key, context_ ? context_->txn_ : nullptr);
                delete[] key;
            }

            // WAL
            RmRecord delete_record{record->size};
            memcpy(delete_record.data, record->data, record->size);
            WriteRecord* write_record = new WriteRecord(WType::DELETE_TUPLE, tab_name_, rid, delete_record);
            context_->txn_->append_write_record(write_record);

            DeleteLogRecord* log = new DeleteLogRecord(context_->txn_->get_transaction_id(),
                                                       context_->txn_->get_prev_lsn(), delete_record, rid, tab_name_);
            context_->log_mgr_->add_log_to_buffer(log);
            context_->txn_->set_prev_lsn(log->lsn_);
            delete log;
        }

        for (int i=0; i<cache->get_cache().size(); i++) {
            auto& acache = cache->get_cache()[i];
            for (int j=0; j<acache.table_list.size(); j++) {
                if (acache.table_list[j] == tab_.name) {
                    cache->get_cache()[i].is_valid = false;
                }
            }
        }
        
        return nullptr;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {}

    Rid& rid() override { return abstract_rid_; }

private:
    TabMeta tab_;                   // 表的元数据
    std::vector<Condition> conds_;  // delete的条件
    RmFileHandle* fh_;              // 表的数据文件句柄
    std::vector<Rid> rids_;         // 需要删除的记录的位置
    std::string tab_name_;          // 表名称
    SmManager* sm_manager_;

    Aggregate_cache* cache;
};