/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "common/common.h"
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_index_scan.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "system/sm.h"

class FilterExecutor : public AbstractExecutor {
public:
    FilterExecutor(std::vector<std::unique_ptr<AbstractExecutor>>&& children, const std::vector<Condition>& conds,
                   SmManager* sm_manager, Context* context, std::string tab_name, int index_cond_num) {
        children_ = std::move(children);
        conds_ = conds;
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = tab_name;
        tab_ = sm_manager_->db_.get_table(tab_name_);
        std::map<CompOp, CompOp> swap_op = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };

        int sub_query_idx = 0;
        int cond_idx = 0;
        for (auto& cond : conds_) {
            if (cond.lhs_col.tab_name != tab_name_) {
                assert(cond.rhs_type == RhsType::COLUMN && cond.rhs_col.tab_name == tab_name_);
                std::swap(cond.lhs_col, cond.rhs_col);
                cond.op = swap_op.at(cond.op);
            }
            if (cond.rhs_type == RhsType::SUB_QUERY) {
                sub_query_map_[cond_idx] = sub_query_idx;
            }
            ++sub_query_idx;
            ++cond_idx;
        }

        for (size_t i = 1; i < children_.size(); ++i) {
            auto& child = children_[i];
            sub_query_col_metas_.push_back(child->cols()[0]);
        }

        index_cond_num_ = index_cond_num;
        stop_ = false;

        auto child = children_[0].get();
        for (auto& cond : conds_) {
            auto left_col = *child->get_col(child->cols(), cond.lhs_col);
            auto left_col_offset = child->get_col_offset(child->cols(), cond.lhs_col);
            cond_lhs_metas_.emplace_back(left_col, left_col_offset);
            if (cond.rhs_type == RhsType::COLUMN) {
                auto right_col = *child->get_col(child->cols(), cond.rhs_col);
                auto right_col_offset = child->get_col_offset(child->cols(), cond.rhs_col);
                cond_rhs_metas_.emplace_back(right_col, right_col_offset);
            }
        }

        record_ = std::make_unique<RmRecord>();
    }

    size_t tupleLen() const override { return children_[0]->tupleLen(); }

    const std::vector<ColMeta>& cols() const override { return children_[0]->cols(); }

    std::string getType() override { return "FilterExecutor"; }

    void beginTuple() override {
        children_[0]->beginTuple();

        for (size_t i = 1; i < children_.size(); ++i) {
            auto& child = children_[i];
            std::vector<std::unique_ptr<RmRecord>> results;
            for (child->beginTuple(); !child->is_end(); child->nextTuple()) {
                results.emplace_back(child->Next());
            }
            sub_query_results_.emplace_back(std::move(results));
        }
        for (auto& [cond_idx, sub_query_idx] : sub_query_map_) {
            auto& cond = conds_[cond_idx];
            auto& sub_query_result = sub_query_results_[sub_query_idx];
            if (cond.op != CompOp::OP_IN && sub_query_result.size() > 1) {
                throw InternalError("Subquery should return only one record");
            }
        }

        // initialize value map
        for (size_t i = 0; i < conds_.size(); ++i) {
            auto& cond = conds_[i];
            size_t cmp_times = 1;
            if (cond.rhs_type == RhsType::LIST) {
                cmp_times = cond.rhs_val_list.size();
            } else if (cond.rhs_type == RhsType::SUB_QUERY) {
                cmp_times = sub_query_results_[sub_query_map_[i]].size();
            }

            value_map_.emplace_back(std::vector<Value>());
            for (size_t j = 0; j < cmp_times; ++j) {
                if (cond.rhs_type == RhsType::VALUE) {
                    value_map_[i].push_back(cond.rhs_val);
                } else if (cond.rhs_type == RhsType::COLUMN) {
                    value_map_[i].push_back(Value());
                } else if (cond.rhs_type == RhsType::LIST) {
                    value_map_[i].push_back(cond.rhs_val_list[j]);
                } else if (cond.rhs_type == RhsType::SUB_QUERY) {
                    auto& sub_query_result = sub_query_results_[sub_query_map_[i]][j];
                    auto& right_col = sub_query_col_metas_[sub_query_map_[i]];
                    Value val;
                    val.ConvertRecordToValue(sub_query_result->data, right_col.type, 0, right_col.len);
                    value_map_[i].push_back(val);
                }
            }
        }

        doNext();
    }

    void nextTuple() override {
        children_[0]->nextTuple();
        doNext();
    }

    std::unique_ptr<RmRecord> Next() override { return children_[0].get()->Next(); }

    void Next(std::unique_ptr<RmRecord>& record) override { children_[0].get()->Next(record); }

    bool is_end() const override { return stop_ || children_[0]->is_end(); }

    Rid& rid() override { return children_[0]->rid(); }

private:
    void doNext() {
        TabMeta tab_;
        auto child = children_[0].get();

        for (; !child->is_end(); child->nextTuple()) {
            child->Next(record_);
            if (!record_) {
                break;
            }

            bool pass = true;
            size_t rhs_idx = 0;
            for (size_t i = 0; i < conds_.size(); ++i) {
                auto& cond = conds_[i];
                size_t cmp_times = 1;
                if (cond.rhs_type == RhsType::LIST) {
                    cmp_times = cond.rhs_val_list.size();
                } else if (cond.rhs_type == RhsType::SUB_QUERY) {
                    cmp_times = sub_query_results_[sub_query_map_[i]].size();
                }

                if (!eval_condition(cond, record_.get(), i, rhs_idx, cmp_times)) {
                    pass = false;
                    stop_ |= ((int)i < index_cond_num_ &&
                              (cond.op == CompOp::OP_NE || cond.op == CompOp::OP_LT || cond.op == CompOp::OP_LE));
                    break;
                }
                if (cond.rhs_type == RhsType::COLUMN) {
                    ++rhs_idx;
                }
            }

            if (pass) {
                break;
            }
        }
    }

    bool eval_condition(const Condition& cond, RmRecord* record, size_t i, size_t j, size_t cmp_times) {
        auto& left_col = cond_lhs_metas_[i].first;
        auto left_offset = cond_lhs_metas_[i].second;
        Value left_val = Value(left_col.type);
        Value right_val = Value();

        if (left_col.type == ColType::TYPE_INT) {
            left_val.int_val = *(int*)(record->data + left_offset);
        } else if (left_col.type == ColType::TYPE_FLOAT) {
            left_val.float_val = *(float*)(record->data + left_offset);
        } else if (left_col.type == ColType::TYPE_STRING) {
            left_val.ConvertRecordToValue(record->data, left_col.type, left_offset, left_col.len);
        }

        bool result = false;
        // TODO: IN 只支持单字段查询
        for (size_t k = 0; k < cmp_times; ++k) {
            Value& right_val_ref = cond.rhs_type == RhsType::COLUMN ? right_val : value_map_[i][k];
            if (cond.rhs_type == RhsType::COLUMN) {
                auto& right_col = cond_rhs_metas_[j].first;
                auto right_offset = cond_rhs_metas_[j].second;
                right_val.ConvertRecordToValue(record->data, right_col.type, right_offset, right_col.len);
            }
            right_val_ref.raw_ptr = (char*)&right_val_ref.int_val;

            switch (cond.op) {
                case CompOp::OP_EQ: {
                    result = left_val == right_val_ref;
                    break;
                }
                case CompOp::OP_NE: {
                    result = left_val != right_val_ref;
                    break;
                }
                case CompOp::OP_LT: {
                    result = left_val < right_val_ref;
                    break;
                }
                case CompOp::OP_GT: {
                    result = left_val > right_val_ref;
                    break;
                }
                case CompOp::OP_LE: {
                    result = left_val <= right_val_ref;
                    break;
                }
                case CompOp::OP_GE: {
                    result = left_val >= right_val_ref;
                    break;
                }
                case CompOp::OP_IN: {
                    result |= left_val == right_val_ref;
                    break;
                }
                default: {
                    throw InternalError("Invalid Condition Compare Operator");
                }
            }
        }

        return result;
    }

private:
    SmManager* sm_manager_;
    std::string tab_name_;
    TabMeta tab_;                                             // 表的元数据
    std::unique_ptr<IndexScanExecutor> index_scan_executor_;  // 索引扫描执行器
    int index_cond_num_;

    std::vector<Condition> conds_;
    std::vector<std::pair<ColMeta, size_t>> cond_lhs_metas_;
    std::vector<std::pair<ColMeta, size_t>> cond_rhs_metas_;

    // filter internal values
    std::unique_ptr<RmRecord> record_;
    std::vector<std::vector<Value>> value_map_;

    // sub query
    bool stop_;
    std::map<int, int> sub_query_map_;
    std::vector<ColMeta> sub_query_col_metas_;
    std::vector<std::vector<std::unique_ptr<RmRecord>>> sub_query_results_;
};
