/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "function/func.h"

// FIXME: now only in-memory version
class GroupbyExecutor : public AbstractExecutor {
public:
    GroupbyExecutor(std::unique_ptr<AbstractExecutor> prev, Context* context, std::vector<TabCol> groupby_cols,
                    std::vector<Condition> having_conds) {
        children_.push_back(std::move(prev));
        context_ = context;
        having_conds_ = having_conds;

        auto prev_cols_ = children_[0]->cols();
        for (auto& col : groupby_cols) {
            auto pos = get_col(prev_cols_, col);
            groupby_cols_.push_back(*pos);
        }

        for (auto& col : having_conds) {
            having_aggr_vals_[col.lhs_col] = nullptr;
            if (col.rhs_type != RhsType::VALUE) {
                having_aggr_vals_[col.rhs_col] = nullptr;
            }
        }

        groupby_len_ = groupby_cols_.back().offset + groupby_cols_.back().len;
        group_idx_ = {0, 0};
    }

    ~GroupbyExecutor() {
        for (auto& key : groupby_keys_) {
            delete[] key;
        }
    }

    size_t tupleLen() const override { return children_[0]->tupleLen(); }

    const std::vector<ColMeta>& cols() const override { return children_[0]->cols(); }

    std::string getType() override { return "GroupbyExecutor"; }

    void beginTuple() override {
        auto& child = children_[0];
        // loop all tuples
        for (child->beginTuple(); !child->is_end(); child->nextTuple()) {
            auto tuple = child->Next();
            char* tuple_key = new char[groupby_len_];

            size_t offset = 0;
            for (auto& col : groupby_cols_) {
                memcpy(tuple_key + offset, tuple->data + col.offset, col.len);
                offset += col.len;
            }

            int idx = -1;
            for (size_t i = 0; i < groupby_keys_.size(); i++) {
                if (memcmp(groupby_keys_[i], tuple_key, groupby_len_) == 0) {
                    idx = (int)i;
                    break;
                }
            }
            if (idx == -1) {
                idx = groupby_keys_.size();
                groupby_keys_.push_back(tuple_key);
                groups_.emplace_back(std::vector<std::unique_ptr<RmRecord>>());
                groups_.back().emplace_back(std::move(tuple));
            } else {
                groups_[idx].emplace_back(std::move(tuple));
                delete[] tuple_key;
            }
        }

        // handle having
        auto it = groups_.begin();
        while (it != groups_.end()) {
            bool erase = false;
            // do aggregate in having
            for (auto& tuple : *it) {
                for (auto& [col, val] : having_aggr_vals_) {
                    ColMeta col_meta;
                    size_t offset = 0;
                    auto pos = get_col(child->cols(), col);
                    if (pos == child->cols().end() && col.aggr_op == AggrOp::AGGR_COUNT && col.col_name == "*") {
                        col_meta.type = ColType::TYPE_INT;
                        col_meta.len = sizeof(int);
                        col_meta.offset = 0;
                    } else {
                        col_meta = *get_col(child->cols(), col);
                        offset = get_col_offset(child->cols(), col);
                    }

                    if (val == nullptr) {
                        val = new char[col_meta.len];
                        if (col.aggr_op == AggrOp::AGGR_COUNT) {
                            *(int32_t*)val = 1;
                        } else {
                            memcpy(val, tuple->data + col_meta.offset, col_meta.len);
                        }
                    } else {
                        switch (col.aggr_op) {
                            case AggrOp::AGGR_COUNT: {
                                Count(col_meta.type, tuple->data + offset, val);
                                break;
                            }
                            case AggrOp::AGGR_SUM: {
                                Sum(col_meta.type, tuple->data + offset, val, col_meta.len);
                                break;
                            }
                            case AggrOp::AGGR_MAX: {
                                Max(col_meta.type, tuple->data + offset, val, col_meta.len);
                                break;
                            }
                            case AggrOp::AGGR_MIN: {
                                Min(col_meta.type, tuple->data + offset, val, col_meta.len);
                                break;
                            }
                            default: {
                                break;
                            }
                        }
                    }
                }
            }

            for (auto& cond : having_conds_) {
                if (!EvalCondition(cond)) {
                    erase = true;
                    break;
                }
            }

            for (auto& [col, val] : having_aggr_vals_) {
                delete[] val;
                val = nullptr;
            }

            if (erase) {
                it = groups_.erase(it);
            } else {
                ++it;
            }
        }
    }

    void nextTuple() override {
        if (++group_idx_.second == groups_[group_idx_.first].size()) {
            group_end_ = true;
            group_idx_.second = 0;
            ++group_idx_.first;
        }
    }

    std::unique_ptr<RmRecord> Next() override { return std::move(groups_[group_idx_.first][group_idx_.second]); }

    void Next(std::unique_ptr<RmRecord>& record) override {
        auto& src = *(groups_[group_idx_.first][group_idx_.second].get());
        record->MoveFrom(src);
    }

    bool is_end() const override { return group_idx_.first == groups_.size(); }

    Rid& rid() override { return abstract_rid_; }

private:
    bool EvalCondition(const Condition& cond) {
        auto child = children_[0].get();
        Value left_val;
        Value right_val;

        char* left_data = having_aggr_vals_[cond.lhs_col];
        if (cond.lhs_col.col_name == "*") {
            left_val.ConvertRecordToValue(left_data, ColType::TYPE_INT, 0, sizeof(int));
        } else {
            auto left_col = *get_col(child->cols(), cond.lhs_col);
            left_val.ConvertRecordToValue(left_data, left_col.type, 0, left_col.len);
        }
        if (cond.rhs_type == RhsType::VALUE) {
            right_val = cond.rhs_val;
        } else if (cond.rhs_type == RhsType::COLUMN) {
            char* right_data = having_aggr_vals_[cond.rhs_col];
            if (cond.rhs_col.col_name == "*") {
                right_val.ConvertRecordToValue(right_data, ColType::TYPE_INT, 0, sizeof(int));
            } else {
                auto right_col = *get_col(child->cols(), cond.rhs_col);
                right_val.ConvertRecordToValue(right_data, right_col.type, 0, right_col.len);
            }
        } else {
            throw InternalError("Invalid Condition Right Hand Side Type");
        }

        bool result = false;
        switch (cond.op) {
            case CompOp::OP_EQ: {
                result = left_val == right_val;
                break;
            }
            case CompOp::OP_NE: {
                result = left_val != right_val;
                break;
            }
            case CompOp::OP_LT: {
                result = left_val < right_val;
                break;
            }
            case CompOp::OP_GT: {
                result = left_val > right_val;
                break;
            }
            case CompOp::OP_LE: {
                result = left_val <= right_val;
                break;
            }
            case CompOp::OP_GE: {
                result = left_val >= right_val;
                break;
            }
            default: {
                throw InternalError("Invalid Condition Compare Operator");
            }
        }
        return result;
    }

private:
    std::vector<ColMeta> groupby_cols_;
    size_t groupby_len_;
    std::vector<char*> groupby_keys_;

    std::vector<Condition> having_conds_;
    std::map<TabCol, char*> having_aggr_vals_;

    std::vector<std::vector<std::unique_ptr<RmRecord>>> groups_;
    std::pair<size_t, size_t> group_idx_;
};