/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "common/config.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_file_handle.h"
#include "system/sm.h"

#include <map>

class IndexScanExecutor : public AbstractExecutor {
public:
    IndexScanExecutor(SmManager* sm_manager, std::string tab_name, std::vector<std::string> index_col_names,
                      std::vector<Condition> conds, int index_cond_num, Context* context) {
        sm_manager_ = sm_manager;
        context_ = context;
        tab_name_ = std::move(tab_name);
        tab_ = sm_manager_->db_.get_table(tab_name_);
        index_col_names_ = index_col_names;
        index_meta_ = *(tab_.get_index_meta(index_col_names_));
        fh_ = sm_manager_->fhs_.at(tab_name_).get();
        cols_ = tab_.cols;
        len_ = cols_.back().offset + cols_.back().len;

        // handle condition
        ih_ = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta_.cols)).get();
        lower_ = GetLowerBound(conds, index_cond_num);
    }

    ~IndexScanExecutor() override {
        for (auto& [_, page] : bpm_) {
            fh_->unpin_page(page, true);
        }
    }

    std::string getType() override { return "IndexScanExecutor"; }

    void beginTuple() override {
        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
        Iid upper = ih_->leaf_end();
        scan_ = std::make_unique<IxScan>(ih_, lower_, upper, sm_manager_->get_bpm());
        rid_ = scan_->rid();
    }

    void nextTuple() override {
        scan_->next();
        if (!scan_->is_end()) {
            rid_ = scan_->rid();
        }
    }

    std::unique_ptr<RmRecord> Next() override {
        if (bpm_.size() > 10000) {
            for (auto& [_, handle] : bpm_) {
                fh_->unpin_page(handle, handle.page->is_dirty());
            }
            bpm_.clear();
        }
        if (bpm_.count(rid_.page_no) == 0) {
            bpm_[rid_.page_no] = fh_->fetch_page_handle(rid_.page_no);
        }
        return fh_->get_record(rid_, context_, bpm_[rid_.page_no]);
        // return fh_->get_record(rid_, context_, bpm_[rid_.page_no]);
    }

    void Next(std::unique_ptr<RmRecord> &record) override {
        if (bpm_.size() > 10000) {
            for (auto& [_, handle] : bpm_) {
                fh_->unpin_page(handle, handle.page->is_dirty());
            }
            bpm_.clear();
        }
        if (bpm_.count(rid_.page_no) == 0) {
            bpm_[rid_.page_no] = fh_->fetch_page_handle(rid_.page_no);
        }
        fh_->get_record(rid_, context_, bpm_[rid_.page_no], record);
    }

    Rid& rid() override { return rid_; }
    bool is_end() const override { return scan_->is_end(); }
    const std::vector<ColMeta>& cols() const override { return cols_; }
    size_t tupleLen() const override { return len_; }

private:
    Iid GetLowerBound(std::vector<Condition>& conds, int index_cond_num) {
        std::map<CompOp, CompOp> swap_map = {
            {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
        };
        for (int i = 0; i < index_cond_num; ++i) {
            auto& cond = conds[i];
            if (cond.lhs_col.tab_name != tab_name_) {
                if (cond.rhs_type == RhsType::COLUMN && cond.rhs_col.tab_name == tab_name_) {
                    std::swap(cond.lhs_col, cond.rhs_col);
                    cond.op = swap_map.at(cond.op);
                }
            }
        }
        bool new_col = true;
        std::vector<Condition> lower_conds;
        for (int i = 0, j = 0; i < index_cond_num; ++i) {
            auto& cond = conds[i];
            if (j < index_col_names_.size() && index_col_names_.at(j) != cond.lhs_col.col_name) {
                ++j;
                new_col = true;
            }
            if (j >= index_col_names_.size()) {
                continue;
            }

            if (cond.op == OP_EQ || cond.op == OP_GE || cond.op == OP_GT) {
                if (new_col) {
                    lower_conds.push_back(cond);
                    new_col = false;
                } else if (lower_conds.back().rhs_val < cond.rhs_val) {
                    lower_conds.back() = cond;
                }
            }
        }

        size_t len = 0;
        for (auto& col : index_meta_.cols) {
            len += col.len;
        }
        char* key = new char[len];

        int i = 0;
        size_t offset = 0;
        for (auto& col : index_meta_.cols) {
            if (i >= lower_conds.size() || lower_conds.at(i).lhs_col.col_name != col.name) {
                // no condition on this column, set to min
                switch (col.type) {
                    case ColType::TYPE_INT: {
                        *(int*)(key + offset) = std::numeric_limits<int>::min();
                        break;
                    }
                    case ColType::TYPE_FLOAT: {
                        *(float*)(key + offset) = std::numeric_limits<float>::lowest();
                        break;
                    }
                    case ColType::TYPE_STRING: {
                        memset(key + offset, 0, col.len);
                        break;
                    }
                }
                --i;
            } else {
                auto& cond = lower_conds.at(i);
                assert(col.name == cond.lhs_col.col_name);
                switch (col.type) {
                    case ColType::TYPE_INT: {
                        *(int*)(key + offset) = cond.rhs_val.int_val;
                        break;
                    }
                    case ColType::TYPE_FLOAT: {
                        *(float*)(key + offset) = cond.rhs_val.float_val;
                        break;
                    }
                    case ColType::TYPE_STRING: {
                        memcpy(key + offset, cond.rhs_val.str_val.c_str(), col.len);
                        break;
                    }
                }
            }
            offset += col.len;
            ++i;
        }

        Iid iid{-1, -1};
        bool equal = false;
        iid = ih_->lower_bound(key, &equal);
        if (iid.page_no == -1 && iid.slot_no == -1) {
            iid = ih_->leaf_begin();
        }
        delete[] key;
        return iid;
    }

private:
    SmManager* sm_manager_;
    std::string tab_name_;       // 表名称
    TabMeta tab_;                // 表的元数据
    std::vector<ColMeta> cols_;  // 需要读取的字段
    size_t len_;                 // 选取出来的一条记录的长度

    std::vector<std::string> index_col_names_;  // index scan涉及到的索引包含的字段
    IndexMeta index_meta_;                      // index scan涉及到的索引元数据

    Rid rid_;
    RmFileHandle* fh_;  // 表的数据文件句柄
    Iid lower_;
    IxIndexHandle* ih_;
    std::unique_ptr<RecScan> scan_;

    std::map<int, RmPageHandle> bpm_;
};