/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <climits>
#include <cstdint>
#include <fstream>
#include <sstream>
#include <string>

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class LoadExecutor : public AbstractExecutor {
private:
    TabMeta tab_;           // 表的元数据
    RmFileHandle* fh_;      // 表的数据文件句柄
    std::string tab_name_;  // 表名称
    std::string path_;
    SmManager* sm_manager_;
    Rid rid_;

    std::ifstream ifs_;
    std::vector<ColMeta> col_metas_;  // 文件中的数据所需要插入的offset

public:
    LoadExecutor(SmManager* sm_manager, const std::string& tab_name, std::string path, Context* context) : ifs_(path) {
        sm_manager_ = sm_manager;
        tab_ = sm_manager_->db_.get_table(tab_name);
        tab_name_ = tab_name;
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        context_ = context;

        readFileHeader();
    };

    std::unique_ptr<RmRecord> Next() override {
        std::string line;
        std::string token;
        std::vector<std::string> data;
        RmRecord rec(fh_->get_file_hdr().record_size);
        char* key = nullptr;
        size_t max_len = 0;

        for (auto& index : tab_.indexes) {
            max_len = index.col_tot_len > max_len ? index.col_tot_len : max_len;
        }
        if (max_len > 0) {
            key = new char[max_len];
        }
        while (std::getline(ifs_, line)) {
            data.clear();
            std::stringstream ss(line);
            while (std::getline(ss, token, ',')) {
                data.push_back(token);
            }

            if (data.size() != col_metas_.size()) {
                throw InternalError("invalid file format");
            }

            for (size_t i = 0; i < col_metas_.size(); i++) {
                auto meta = col_metas_.at(i);
                token = data.at(i);
                switch (meta.type) {
                    case TYPE_INT: {
                        *(int*)(rec.data + meta.offset) = std::stoi(token);
                        break;
                    }
                    case TYPE_FLOAT: {
                        *(float*)(rec.data + meta.offset) = std::stof(token);
                        break;
                    }
                    case TYPE_STRING: {
                        memcpy(rec.data + meta.offset, token.c_str(),
                               std::min((size_t)meta.len, strlen(token.c_str())));
                        break;
                    }
                    default:
                        break;
                }
            }

            // Insert into record file
            Rid id = fh_->insert_record(rec.data, nullptr);

            // Insert into index
            for (size_t i = 0; i < tab_.indexes.size(); ++i) {
                auto& index = tab_.indexes[i];
                auto ih =
                    sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
                int offset = 0;
                for (size_t i = 0; i < index.col_num; ++i) {
                    memcpy(key + offset, rec.data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->insert_entry(key, id, context_->txn_);
            }
        }

        if (key != nullptr) {
            delete[] key;
        }

        ifs_.close();
        return nullptr;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {}

    Rid& rid() override { return rid_; }

private:
    void readFileHeader() {
        // 读取数据文件的列名
        if (!ifs_.is_open()) {
            throw UnixError();
        }

        std::string line;
        std::getline(ifs_, line);

        std::string token;
        std::vector<std::string> cols;
        std::stringstream ss(line);
        while (std::getline(ss, token, ',')) {
            cols.push_back(token);
        }

        for (auto& col : cols) {
            auto it = std::find_if(tab_.cols.begin(), tab_.cols.end(),
                                   [&](const ColMeta& a) -> bool { return a.name == col; });
            col_metas_.push_back(*it);
        }
    }
};