/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class NestedLoopJoinExecutor : public AbstractExecutor {
public:
    NestedLoopJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right,
                           std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto& col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        fed_conds_ = std::move(conds);

        outer_end_ = true;
        inner_end_ = false;
        outer_idx_ = 0;
        inner_idx_ = 0;
    }

    size_t tupleLen() const override { return len_; }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    std::string getType() override { return "NestedLoopJoinExecutor"; }

    void beginTuple() override {
        right_->beginTuple();
        left_->beginTuple();
        doNext(0, 0);
    }

    void nextTuple() override { doNext(1, -1); }

    std::unique_ptr<RmRecord> Next() override {
        auto& left_cols = left_->cols();
        auto& right_cols = right_->cols();
        char* data = new char[len_];
        char* left_data = inner_records_.at(inner_idx_)->data;
        char* right_data = outer_record_page_.at(outer_idx_)->data;
        size_t offset = 0;

        for (auto& col : cols_) {
            auto left_pos = std::find_if(left_cols.begin(), left_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            auto right_pos = std::find_if(right_cols.begin(), right_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            if (left_pos == left_cols.end() && right_pos == right_cols.end()) {
                throw ColumnNotFoundError(col.tab_name + '.' + col.name);
            }

            if (left_pos != left_cols.end()) {
                memcpy(data + offset, left_data + left_pos->offset, left_pos->len);
            }
            if (right_pos != right_cols.end()) {
                memcpy(data + offset, right_data + right_pos->offset, right_pos->len);
            }

            offset += col.len;
        }

        assert(offset == len_);
        // FIXME: remove memcpy
        auto result = std::make_unique<RmRecord>(len_, data);
        delete[] data;
        return result;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {
        auto& left_cols = left_->cols();
        auto& right_cols = right_->cols();
        char* data = new char[len_];
        char* left_data = inner_records_.at(inner_idx_)->data;
        char* right_data = outer_record_page_.at(outer_idx_)->data;
        size_t offset = 0;

        for (auto& col : cols_) {
            auto left_pos = std::find_if(left_cols.begin(), left_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            auto right_pos = std::find_if(right_cols.begin(), right_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            if (left_pos == left_cols.end() && right_pos == right_cols.end()) {
                throw ColumnNotFoundError(col.tab_name + '.' + col.name);
            }

            if (left_pos != left_cols.end()) {
                memcpy(data + offset, left_data + left_pos->offset, left_pos->len);
            }
            if (right_pos != right_cols.end()) {
                memcpy(data + offset, right_data + right_pos->offset, right_pos->len);
            }

            offset += col.len;
        }

        assert(offset == len_);
        if (record->allocated_) {
            delete[] record->data;
        }
        record->data = data;
        record->size = len_;
        record->allocated_ = true;
    }

    bool is_end() const override { return right_->is_end() && outer_end_ && outer_record_page_.empty(); }

    Rid& rid() override { return outer_rid_; }

private:
    void doNext(size_t inner_delta, size_t inner_pos) {
        while (!is_end()) {
            // Read a page of record from right table
            if (outer_end_) {
                for (; !right_->is_end(); right_->nextTuple()) {
                    if (outer_end_) {
                        outer_end_ = false;
                        outer_rid_ = right_->rid();
                    }
                    if (right_->rid().page_no == outer_rid_.page_no) {
                        outer_record_page_.push_back(right_->Next());
                        continue;
                    }
                    break;
                }
            }

            // Join a page of right table with all records in left table
            for (; outer_idx_ != outer_record_page_.size(); ++outer_idx_) {
                while (!inner_end_) {
                    if (inner_records_.empty()) {
                        std::unordered_map<int, int> map;
                        for (; !left_->is_end(); left_->nextTuple()) {
                            if (map.count(left_->rid().page_no) == 0 && map.size() == BUFFER_POOL_SIZE - 5000) {
                                break;
                            }
                            map[left_->rid().page_no] = 1;
                            inner_records_.push_back(left_->Next());
                        }
                        map.clear();
                    }

                    bool ok = false;
                    inner_idx_ += inner_delta;
                    for (; inner_idx_ != inner_records_.size(); ++inner_idx_) {
                        ok = true;
                        for (auto& cond : fed_conds_) {
                            if (!eval_condition(cond, inner_records_.at(inner_idx_).get(),
                                                outer_record_page_.at(outer_idx_).get())) {
                                ok = false;
                                break;
                            }
                        }
                        if (ok) {
                            return;
                        }
                    }

                    inner_records_.clear();
                    inner_idx_ = inner_pos;
                    if (left_->is_end()) {
                        inner_end_ = true;
                    }
                }
                left_->beginTuple();
                inner_end_ = false;
            }

            outer_record_page_.clear();
            outer_idx_ = 0;
            outer_end_ = true;
            left_->beginTuple();
        }
    }

    bool eval_condition(const Condition& cond, RmRecord* left, RmRecord* right) {
        auto left_col = left_->get_col(left_->cols(), cond.lhs_col);
        auto left_offset = left_->get_col_offset(left_->cols(), cond.lhs_col);
        Value left_val = Value();
        Value right_val = Value();

        left_val.ConvertRecordToValue(left->data, left_col->type, left_offset, left_col->len);
        if (cond.rhs_type == RhsType::VALUE) {
            right_val = cond.rhs_val;
        } else if (cond.rhs_type == RhsType::COLUMN) {
            auto right_col = right_->get_col(right_->cols(), cond.rhs_col);
            auto right_offset = right_->get_col_offset(right_->cols(), cond.rhs_col);
            right_val.ConvertRecordToValue(right->data, right_col->type, right_offset, right_col->len);
        } else {
            throw InternalError("Invalid Condition Right Hand Side Type");
        }

        bool result = false;
        switch (cond.op) {
            case CompOp::OP_EQ: {
                result = left_val == right_val;
                break;
            }
            case CompOp::OP_NE: {
                result = left_val != right_val;
                break;
            }
            case CompOp::OP_LT: {
                result = left_val < right_val;
                break;
            }
            case CompOp::OP_GT: {
                result = left_val > right_val;
                break;
            }
            case CompOp::OP_LE: {
                result = left_val <= right_val;
                break;
            }
            case CompOp::OP_GE: {
                result = left_val >= right_val;
                break;
            }
            default: {
                throw InternalError("Invalid Condition Compare Operator");
            }
        }
        return result;
    }

private:
    std::unique_ptr<AbstractExecutor> left_;   // 左儿子节点（需要join的表）
    std::unique_ptr<AbstractExecutor> right_;  // 右儿子节点（需要join的表）
    size_t len_;                               // join后获得的每条记录的长度
    std::vector<ColMeta> cols_;                // join后获得的记录的字段

    std::vector<Condition> fed_conds_;  // join条件

    bool outer_end_;
    bool inner_end_;
    Rid outer_rid_;
    size_t outer_idx_;
    size_t inner_idx_;

    std::vector<std::unique_ptr<RmRecord>> outer_record_page_;
    std::vector<std::unique_ptr<RmRecord>> inner_records_;
};