/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record_printer.h"
#include "system/sm.h"

#include <cstring>

class PrintExecutor : public AbstractExecutor {
public:
    PrintExecutor(std::unique_ptr<AbstractExecutor> prev, std::string file_name, Context* context) : ss_(result_) {
        children_.emplace_back(std::move(prev));
        context_ = context;
    }

    size_t tupleLen() const override { return children_[0]->tupleLen(); }

    const std::vector<ColMeta>& cols() const override { return children_[0]->cols(); }

    std::string getType() override { return "PrintExecutor"; }

    void beginTuple() override {
        children_[0]->beginTuple();
        auto& cols = children_[0]->cols();

        std::vector<std::string> captions;
        captions.reserve(cols.size());
        for (auto& sel_col : cols) {
            captions.emplace_back(sel_col.name);
        }
        ss_ << "|";
        for (auto& caption : captions) {
            ss_ << caption << "|";
        }
        ss_ << "\n";
    }

    void nextTuple() override { children_[0]->nextTuple(); }

    std::unique_ptr<RmRecord> Next() override {
        auto record = children_[0]->Next();
        auto& cols = children_[0]->cols();

        ss_ << "|";
        for (auto& col : cols) {
            char* rec_buf = record->data + col.offset;
            if (col.type == TYPE_INT) {
                ss_ << std::to_string(*(int*)rec_buf) << "|";
            } else if (col.type == TYPE_FLOAT) {
                ss_ << std::to_string(*(float*)rec_buf) << "|";
            } else if (col.type == TYPE_STRING) {
                auto str = std::string((char*)rec_buf, col.len);
                str.resize(strlen(str.c_str()));
                ss_ << str << "|";
            }
        }
        ss_ << "\n";

        return record;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {
        children_[0]->Next(record);
        auto& cols = children_[0]->cols();

        ss_ << "|";
        for (auto& col : cols) {
            char* rec_buf = record->data + col.offset;
            if (col.type == TYPE_INT) {
                ss_ << std::to_string(*(int*)rec_buf) << "|";
            } else if (col.type == TYPE_FLOAT) {
                ss_ << std::to_string(*(float*)rec_buf) << "|";
            } else if (col.type == TYPE_STRING) {
                auto str = std::string((char*)rec_buf, col.len);
                str.resize(strlen(str.c_str()));
                ss_ << str << "|";
            }
        }
        ss_ << "\n";
    }

    void Print() {
        sorted_file_.open("sorted_results.txt", std::ios::out | std::ios::app);
        sorted_file_ << ss_.str();
        sorted_file_.close();
    }

    bool is_end() const override { return children_[0]->is_end(); }

    Rid& rid() override { return children_[0]->rid(); }

private:
    // only for test
    std::fstream sorted_file_;
    std::string result_;
    std::stringstream ss_;
};