/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class ProjectionExecutor : public AbstractExecutor {
public:
    ProjectionExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol>& sel_cols) {
        prev_ = std::move(prev);
        // if (!prev_ || cols_.empty()) {
        //     throw std::runtime_error("Invalid arguments to ProjectionExecutor constructor");
        // }
        size_t curr_offset = 0;
        auto& prev_cols = prev_->cols();
        for (auto& sel_col : sel_cols) {
            auto pos = get_col(prev_cols, sel_col);
            if (pos == prev_cols.end() && sel_col.aggr_op == AggrOp::AGGR_COUNT && sel_col.col_name == "*") {
                ColMeta col_meta = {.tab_name = sel_col.tab_name,
                                    .name = sel_col.col_name,
                                    .type = ColType::TYPE_INT,
                                    .len = sizeof(int),
                                    .offset = (int)curr_offset};
                col_meta.aggr_op = AggrOp::AGGR_COUNT;
                sel_idxs_.push_back(prev_cols.size());
                cols_.push_back(col_meta);
                curr_offset += sizeof(int);
                continue;
            }

            sel_idxs_.push_back(pos - prev_cols.begin());
            auto col = *pos;
            col.offset = curr_offset;
            curr_offset += col.len;
            cols_.push_back(col);
        }
        len_ = curr_offset;
    }

    size_t tupleLen() const override { return len_; }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    void beginTuple() override { prev_->beginTuple(); }

    void nextTuple() override { prev_->nextTuple(); }

    std::unique_ptr<RmRecord> Next() override {
        auto record = prev_->Next();
        if (!record) {
            return nullptr;
        }

        auto data = record->data;
        auto new_data = new char[len_];
        for (size_t i = 0; i < sel_idxs_.size(); i++) {
            // COUNT STAR
            if (sel_idxs_[i] == prev_->cols().size()) {
                int count = 0;
                memcpy(new_data + cols_[i].offset, &count, sizeof(int));
                continue;
            }

            auto& col = cols_[i];
            auto& prev_col = prev_->cols()[sel_idxs_[i]];
            memcpy(new_data + col.offset, data + prev_col.offset, col.len);
        }
        record->ResetData(new_data);
        return record;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {
        prev_->Next(record);

        auto data = record->data;
        auto new_data = new char[len_];
        for (size_t i = 0; i < sel_idxs_.size(); i++) {
            // COUNT STAR
            if (sel_idxs_[i] == prev_->cols().size()) {
                int count = 0;
                memcpy(new_data + cols_[i].offset, &count, sizeof(int));
                continue;
            }

            auto& col = cols_[i];
            auto& prev_col = prev_->cols()[sel_idxs_[i]];
            memcpy(new_data + col.offset, data + prev_col.offset, col.len);
        }
        record->ResetData(new_data);
    }

    bool is_end() const override { return prev_->is_end(); }

    Rid& rid() override { return abstract_rid_; }

private:
    // The previous operator
    std::unique_ptr<AbstractExecutor> prev_;
    // The columns to be projected
    std::vector<ColMeta> cols_;
    // The length of the record after projection
    size_t len_;
    // The indexes of the selected columns in the previous operator
    std::vector<size_t> sel_idxs_;
};