/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SeqScanExecutor : public AbstractExecutor {
public:
    SeqScanExecutor(SmManager* sm_manager, std::string tab_name, std::vector<Condition> conds, Context* context)
        : sm_manager_(sm_manager),
          tab_name_(std::move(tab_name)),
          fh_(sm_manager_->fhs_.at(tab_name_).get()),
          scan_(std::make_unique<RmScan>(fh_)),
          rid_(scan_->rid()) {
        context_ = context;

        TabMeta& tab = sm_manager_->db_.get_table(tab_name_);
        cols_ = tab.cols;
        len_ = cols_.back().offset + cols_.back().len;
        scan_->init(conds, cols_);

        page_handle_ = {-1, RmPageHandle()};
    }

    ~SeqScanExecutor() override {
        if (page_handle_.first != -1) {
            auto& handle = page_handle_.second;
            fh_->unpin_page(handle, handle.page->is_dirty());
        }
    }

    size_t tupleLen() const override { return len_; }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    std::string getType() override { return "SeqScanExecutor"; }

    void beginTuple() override {
        context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
        scan_->begin();
    }

    void nextTuple() override { scan_->next(); }

    std::unique_ptr<RmRecord> Next() override {
        if (page_handle_.first != rid_.page_no) {
            if (page_handle_.first != -1) {
                auto& handle = page_handle_.second;
                fh_->unpin_page(handle, handle.page->is_dirty());
            }
            page_handle_ = {rid_.page_no, fh_->fetch_page_handle(rid_.page_no)};
        }
        return fh_->get_record(rid_, context_, page_handle_.second);
        // return fh_->get_record(rid_, context_);
    }

    void Next(std::unique_ptr<RmRecord>& record) override {
        if (page_handle_.first != rid_.page_no) {
            if (page_handle_.first != -1) {
                auto& handle = page_handle_.second;
                fh_->unpin_page(handle, handle.page->is_dirty());
            }
            page_handle_ = {rid_.page_no, fh_->fetch_page_handle(rid_.page_no)};
        }
        fh_->get_record(rid_, context_, page_handle_.second, record);
    }

    bool is_end() const override { return scan_->is_end(); }

    Rid& rid() override { return rid_; }

private:
    SmManager* sm_manager_;
    std::string tab_name_;       // 表的名称
    RmFileHandle* fh_;           // 表的数据文件句柄
    std::vector<ColMeta> cols_;  // scan后生成的记录的字段
    size_t len_;                 // scan后生成的每条记录的长度

    std::unique_ptr<RmScan> scan_;  // table_iterator
    Rid& rid_;

    std::pair<int, RmPageHandle> page_handle_;
};