/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

#include <queue>
#include <thread>
#include <chrono>

class SortExecutor : public AbstractExecutor {
public:
    static constexpr int kSortThreads = 8;

public:
    SortExecutor(std::unique_ptr<AbstractExecutor> prev, const std::vector<TabCol>& order_cols,
                 std::vector<bool> is_desc, int limit)
        : cmp_(this), pq_(cmp_) {
        prev_ = std::move(prev);
        is_desc_ = is_desc;
        limit_ = limit;
        total_num_ = 0;
        tuple_idx_ = 0;
        for (auto col = order_cols.begin(); col != order_cols.end(); col++) {
            cols_.push_back(*prev_->get_col(prev_->cols(), *col));
        }
        for (int i = 0; i < kSortThreads; ++i) {
            sorted_tuple_[i].clear();
            sorted_idx_.push_back(0);
        }
    }

    size_t tupleLen() const override { return prev_->tupleLen(); }

    const std::vector<ColMeta>& cols() const override { return prev_->cols(); }

    std::string getType() override { return "SortExecutor"; }

    void beginTuple() override {
        std::this_thread::sleep_for(std::chrono::milliseconds(500));
        int i = 0;
        prev_->beginTuple();
        for (; !prev_->is_end(); prev_->nextTuple()) {
            auto record = prev_->Next();
            sorted_tuple_[i].emplace_back(std::move(record));
            if (sorted_tuple_[i].size() >= 1) {
                i++;
                i %= kSortThreads;
            }
        }

        std::vector<std::thread> ths;
        for (int j = 0; j < kSortThreads; ++j) {
            std::thread th([&, j]() { Sort(sorted_tuple_[j]); });
            ths.push_back(std::move(th));
        }
        for (auto& th : ths) {
            th.join();
        }

        // build priority_queue
        for (int j = 0; j < kSortThreads; ++j) {
            if (sorted_tuple_[j].empty()) {
                continue;
            }
            pq_.push({{j, 0}, sorted_tuple_[j][0].get()});
            // ++sorted_idx_[j];
        }

        for (int j = 0; j < kSortThreads; ++j) {
            total_num_ += sorted_tuple_[j].size();
        }
        tuple_idx_ = 0;
    }

    void nextTuple() override {
        int idx = pq_.top().first.first;
        int cnt = 0;
        pq_.pop();
        while (++sorted_idx_[idx] >= sorted_tuple_[idx].size() && cnt < kSortThreads) {
            ++idx;
            ++cnt;
            idx %= kSortThreads;
        }
        if (sorted_idx_[idx] < sorted_tuple_[idx].size()) {
            pq_.push({{idx, sorted_idx_[idx]}, sorted_tuple_[idx][sorted_idx_[idx]].get()});
        }
        tuple_idx_++;
    }

    std::unique_ptr<RmRecord> Next() override {
        auto idx = pq_.top().first;
        return std::move(sorted_tuple_[idx.first][idx.second]);
    }

    void Next(std::unique_ptr<RmRecord>& record) override {
        auto idx = pq_.top().first;
        auto &src = *(sorted_tuple_[idx.first][idx.second]).get();
        record->MoveFrom(src);
    }

    bool is_end() const override { return tuple_idx_ >= total_num_ || (limit_ > 0 && tuple_idx_ >= limit_); }

    Rid& rid() override { return abstract_rid_; }

private:
    void TestResultPrintHeader() {}

    void TestResultPrintRecord() {}

    void Sort(std::vector<std::unique_ptr<RmRecord>>& tuples) {
        Comparator cmp(this);
        std::stable_sort(tuples.begin(), tuples.end(),
                         [&cmp](const std::unique_ptr<RmRecord>& a, const std::unique_ptr<RmRecord>& b) {
                             return cmp(a.get(), b.get());
                         });
    }

private:
    using Pair = std::pair<std::pair<int, int>, RmRecord*>;

    struct Comparator {
        SortExecutor* sort_executor;
        Comparator(SortExecutor* sort_executor) : sort_executor(sort_executor) {}

        bool operator()(const Pair& a, const Pair& b) { return !this->operator()(a.second, b.second); }

        bool operator()(const RmRecord* a, const RmRecord* b) {
            bool a_prev = false;
            for (auto item = sort_executor->cols_.begin(); item != sort_executor->cols_.end(); item++) {
                char* a_ptr = a->data + item->offset;
                char* b_ptr = b->data + item->offset;
                int cmp = ix_compare(a_ptr, b_ptr, item->type, item->len);
                if (cmp < 0) {
                    bool is_order_desc = sort_executor->is_desc_[item - sort_executor->cols_.begin()];
                    if (!is_order_desc) {
                        a_prev = true;
                    }
                    break;
                } else if (cmp > 0) {
                    bool is_order_desc = sort_executor->is_desc_[item - sort_executor->cols_.begin()];
                    if (is_order_desc) {
                        a_prev = true;
                    }
                    break;
                } else {
                }
            }
            return a_prev;
        }
    };

private:
    std::unique_ptr<AbstractExecutor> prev_;
    std::vector<ColMeta> cols_;  // 框架中只支持一个键排序，需要自行修改数据结构支持多个键排序
    std::vector<bool> is_desc_;  // 是否降序
    int limit_;                  // 最大返回数量限制

    std::vector<std::unique_ptr<RmRecord>> sorted_tuple_[kSortThreads];  // 排序过的tuple
    std::vector<int> sorted_idx_;                                        // 当前排序的tuple索引
    int total_num_;
    int tuple_idx_;

    Comparator cmp_;
    std::priority_queue<Pair, std::vector<Pair>, Comparator> pq_;
};