/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "executor_print.h"
#include "index/ix.h"
#include "system/sm.h"

#include <memory>

class SortMergeJoinExecutor : public AbstractExecutor {
public:
    SortMergeJoinExecutor(std::unique_ptr<AbstractExecutor> left, std::unique_ptr<AbstractExecutor> right,
                          std::vector<Condition> conds) {
        left_ = std::move(left);
        right_ = std::move(right);
        len_ = left_->tupleLen() + right_->tupleLen();
        cols_ = left_->cols();
        auto right_cols = right_->cols();
        for (auto& col : right_cols) {
            col.offset += left_->tupleLen();
        }

        cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
        fed_conds_ = std::move(conds);

        left_record_ = nullptr;
        right_record_ = nullptr;
    }

    size_t tupleLen() const override { return len_; }

    const std::vector<ColMeta>& cols() const override { return cols_; }

    std::string getType() override { return "SortMergeJoinExecutor"; }

    void beginTuple() override {
        right_->beginTuple();
        left_->beginTuple();
        DoNext();
    }

    void nextTuple() override {
        left_->nextTuple();
        left_record_ = nullptr;
        DoNext();
    }

    std::unique_ptr<RmRecord> Next() override {
        auto& left_cols = left_->cols();
        auto& right_cols = right_->cols();
        char* data = new char[len_];
        char* left_data = left_record_->data;
        char* right_data = right_record_->data;
        size_t offset = 0;

        for (auto& col : cols_) {
            auto left_pos = std::find_if(left_cols.begin(), left_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            if (left_pos != left_cols.end()) {
                memcpy(data + offset, left_data + left_pos->offset, left_pos->len);
                offset += col.len;
                continue;
            }
            auto right_pos = std::find_if(right_cols.begin(), right_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            if (right_pos != right_cols.end()) {
                memcpy(data + offset, right_data + right_pos->offset, right_pos->len);
                offset += col.len;
                continue;
            }

            throw ColumnNotFoundError(col.tab_name + '.' + col.name);
        }

        assert(offset == len_);
        // FIXME: remove memcpy
        auto result = std::make_unique<RmRecord>(len_, data);
        delete[] data;
        return result;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {
        auto& left_cols = left_->cols();
        auto& right_cols = right_->cols();
        char* data = new char[len_];
        char* left_data = left_record_->data;
        char* right_data = right_record_->data;
        size_t offset = 0;

        for (auto& col : cols_) {
            auto left_pos = std::find_if(left_cols.begin(), left_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            if (left_pos != left_cols.end()) {
                memcpy(data + offset, left_data + left_pos->offset, left_pos->len);
                offset += col.len;
                continue;
            }
            auto right_pos = std::find_if(right_cols.begin(), right_cols.end(), [&](const ColMeta& c) {
                return c.tab_name == col.tab_name && c.name == col.name;
            });
            if (right_pos != right_cols.end()) {
                memcpy(data + offset, right_data + right_pos->offset, right_pos->len);
                offset += col.len;
                continue;
            }

            throw ColumnNotFoundError(col.tab_name + '.' + col.name);
        }

        assert(offset == len_);
        if (record->allocated_) {
            delete[] record->data;
        }
        record->data = data;
        record->size = len_;
        record->allocated_ = true;
    }

    bool is_end() const override {
        bool end = left_->is_end() || right_->is_end();
        if (end) {
            if (auto p = dynamic_cast<PrintExecutor*>(right_.get())) {
                p->Print();
            }
            if (auto p = dynamic_cast<PrintExecutor*>(left_.get())) {
                p->Print();
            }
        }
        return end;
    }

    Rid& rid() override { return abstract_rid_; }

private:
    // FIXME: replace with void Next()
    void DoNext() {
        while (true) {
            if (left_->is_end() || right_->is_end()) {
                break;
            }

            if (!left_record_) {
                left_record_ = left_->Next();
            }
            if (!right_record_) {
                right_record_ = right_->Next();
            }
            int cmp = 0;
            for (auto& cond : fed_conds_) {
                cmp = EvalCondition(cond);
                if (cmp != 0) {
                    break;
                }
            }
            if (cmp == 0) {
                break;
            } else if (cmp < 0) {
                left_->nextTuple();
                left_record_ = nullptr;
            } else {
                right_->nextTuple();
                right_record_ = nullptr;
            }
        }
    }

    int EvalCondition(const Condition& cond) {
        auto left_col = left_->get_col(left_->cols(), cond.lhs_col);
        auto left_offset = left_->get_col_offset(left_->cols(), cond.lhs_col);
        auto right_col = right_->get_col(right_->cols(), cond.rhs_col);
        auto right_offset = right_->get_col_offset(right_->cols(), cond.rhs_col);

        Value left_val = Value();
        Value right_val = Value();
        left_val.ConvertRecordToValue(left_record_->data, left_col->type, left_offset, left_col->len);
        right_val.ConvertRecordToValue(right_record_->data, right_col->type, right_offset, left_col->len);

        return left_val < right_val ? -1 : (left_val == right_val ? 0 : 1);
    }

private:
    std::unique_ptr<AbstractExecutor> left_;
    std::unique_ptr<AbstractExecutor> right_;  // 大表
    size_t len_;
    std::vector<ColMeta> cols_;

    std::vector<Condition> fed_conds_;

    std::unique_ptr<RmRecord> left_record_;
    std::unique_ptr<RmRecord> right_record_;
};