/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
public:
    UpdateExecutor(SmManager* sm_manager, const std::string& tab_name, std::vector<SetClause> set_clauses,
                   std::vector<Condition> conds, std::vector<Rid> rids, Context* context, Aggregate_cache* cache_) {
        sm_manager_ = sm_manager;
        tab_name_ = tab_name;
        set_clauses_ = set_clauses;
        tab_ = sm_manager_->db_.get_table(tab_name);
        fh_ = sm_manager_->fhs_.at(tab_name).get();
        conds_ = conds;
        rids_ = rids;
        context_ = context;

        cache = cache_;
    }

    std::string getType() override { return "UpdateExecutor"; }

    std::unique_ptr<RmRecord> Next() override {
        size_t record_len = tab_.cols.back().offset + tab_.cols.back().len;
        std::vector<std::pair<IndexMeta, IxIndexHandle*>> index_map;
        char* new_data = new char[record_len];
        for (auto& index : tab_.indexes) {
            index_map.emplace_back(
                index,
                sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get());
        }

        for (auto rid : rids_) {
            auto record = fh_->get_record(rid, context_);
            auto data = record->data;

            memcpy(new_data, data, record_len);

            // record a update operation into the transaction
            RmRecord update_record{record->size, data};
            // 添加事务的writeset(update类型)
            WriteRecord* write_record = new WriteRecord(WType::UPDATE_TUPLE, tab_name_, rid, update_record);
            context_->txn_->append_write_record(write_record);

            for (auto& clause : set_clauses_) {
                auto& col = *tab_.get_col(clause.lhs.col_name);
                auto& val = clause.rhs;
                if (col.type != val.type && !(col.type != ColType::TYPE_STRING && val.type != ColType::TYPE_STRING)) {
                    throw IncompatibleTypeError(coltype2str(col.type), coltype2str(val.type));
                }
                if (col.type == ColType::TYPE_INT && val.type == ColType::TYPE_FLOAT) {
                    val.set_int((int)val.float_val);
                }
                if (col.type == ColType::TYPE_FLOAT && val.type == ColType::TYPE_INT) {
                    val.set_float((float)val.int_val);
                }
                val.init_raw(col.len);
                // memcpy(new_data + col.offset, val.raw->data, col.len);
                if (clause.rhs_has_lhs == false) memcpy(new_data + col.offset, val.raw->data, col.len);
                else if (clause.operation == 1){
                    // std::cout << "in add" << std::endl;
                    auto& col_rhs = *tab_.get_col(clause.rhs_col.col_name);
                    if (col.type != col_rhs.type && !(col.type != ColType::TYPE_STRING && col_rhs.type != ColType::TYPE_STRING)) {
                        throw IncompatibleTypeError(coltype2str(col.type), coltype2str(col_rhs.type));
                    }
                    if (col.type == ColType::TYPE_INT) {
                        Value new_value;
                        int record_int;
                        if (col_rhs.type == ColType::TYPE_INT) record_int = *reinterpret_cast<int*>(data + col_rhs.offset);
                        else if (col_rhs.type == ColType::TYPE_FLOAT) {
                            float record_float = *reinterpret_cast<float*>(data + col.offset);
                            record_int = (int)record_float;
                        }
                        new_value.set_int(val.int_val+record_int);
                        new_value.init_raw(col.len);
                        memcpy(new_data + col.offset, new_value.raw->data, col.len);
                    }
                    else if (col.type == ColType::TYPE_FLOAT) {
                        Value new_value;
                        float record_float;
                        if (col_rhs.type == ColType::TYPE_FLOAT) record_float = *reinterpret_cast<float*>(data + col.offset);
                        else if (col_rhs.type == ColType::TYPE_INT) {
                            int record_int = *reinterpret_cast<int*>(data + col_rhs.offset);
                            record_float = (float)record_int;
                        }
                        new_value.set_float(val.float_val+record_float);
                        new_value.init_raw(col.len);
                        // std::cout << "Float_val: " << new_value.float_val << std::endl; 
                        memcpy(new_data + col.offset, new_value.raw->data, col.len);
                    }
                    else memcpy(new_data + col.offset, val.raw->data, col.len);
                }              
            }

            if (!CheckUniqueness(index_map, data, new_data)) {
                delete[] new_data;
                throw DuplicateIndexKeyError();
            }

            fh_->update_record(rid, new_data, context_);

            for (auto& [index, ih] : index_map) {
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for (size_t i = 0; i < index.col_num; ++i) {
                    memcpy(key + offset, data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->delete_entry(key, context_ ? context_->txn_ : nullptr);
                offset = 0;
                for (size_t i = 0; i < index.col_num; ++i) {
                    memcpy(key + offset, new_data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->insert_entry(key, rid, context_ ? context_->txn_ : nullptr);
                delete[] key;
            }
        }
        delete[] new_data;

        for (int i=0; i<cache->get_cache().size(); i++) {
            auto& acache = cache->get_cache()[i];
            for (int j=0; j<acache.table_list.size(); j++) {
                if (acache.table_list[j] == tab_.name) {
                    cache->get_cache()[i].is_valid = false;
                }
            }
        }
        
        return nullptr;
    }

    void Next(std::unique_ptr<RmRecord>& record) override {}

    Rid& rid() override { return abstract_rid_; }

private:
    bool CheckUniqueness(const std::vector<std::pair<IndexMeta, IxIndexHandle*>>& index_map, char* old_data,
                         char* data) {
        bool unique = true;
        bool equal = true;
        std::vector<Rid> rids;
        for (size_t i = 0; i < index_map.size() && unique; ++i) {
            auto& [index, ih] = index_map[i];
            char* key = new char[index.col_tot_len];
            int offset = 0;
            for (size_t i = 0; i < index.col_num; ++i) {
                memcpy(key + offset, data + index.cols[i].offset, index.cols[i].len);
                equal &= memcmp(old_data + index.cols[i].offset, data + index.cols[i].offset, index.cols[i].len) == 0;
                offset += index.cols[i].len;
            }
            unique = equal || !ih->get_value(key, &rids, context_ ? context_->txn_ : nullptr);
            delete[] key;
        }
        return unique;
    }

private:
    TabMeta tab_;
    std::vector<Condition> conds_;
    RmFileHandle* fh_;
    std::vector<Rid> rids_;
    std::string tab_name_;
    std::vector<SetClause> set_clauses_;
    SmManager* sm_manager_;

    Aggregate_cache* cache;
};