/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include <atomic>
#include <condition_variable>
#include <iostream>
#include <mutex>
#include <queue>
#include <thread>

#include "analyze/analyze.h"
#include "errors.h"
#include "optimizer/optimizer.h"
#include "optimizer/plan.h"
#include "optimizer/planner.h"
#include "portal.h"
#include "recovery/log_recovery.h"

class Server {
public:
    Server(int fd) : fd_(fd), exit_(false), read_wait_(0), sql_wait_(0), read_(0), write_(0) {}

    ~Server() = default;

    void init(LockManager* lock_manager, TransactionManager* txn_manager, Optimizer* optimizer, QlManager* ql_manager,
              LogManager* log_manager, Portal* portal, Analyze* analyze, pthread_mutex_t* buffer_mutex, Aggregate_cache* cache_) {
        lock_manager_ = lock_manager;
        txn_manager_ = txn_manager;
        optimizer_ = optimizer;
        ql_manager_ = ql_manager;
        log_manager_ = log_manager;
        portal_ = portal;
        analyze_ = analyze;
        buffer_mutex_ = buffer_mutex;

        cache = cache_;
    }

    void start() {
        read_thread_ = std::thread(&Server::read_data, this);
        sql_thread_ = std::thread(&Server::process_sql, this);
    }

    void wait() {
        read_thread_.join();
        sql_thread_.join();

        std::cout << "read_wait: " << read_wait_ << std::endl;
        std::cout << "sql_wait: " << sql_wait_ << std::endl;
    }

private:
    void set_transaction(txn_id_t* txn_id, Context* context) {
        context->txn_ = txn_manager_->get_transaction(*txn_id);
        if (context->txn_ == nullptr || context->txn_->get_state() == TransactionState::COMMITTED ||
            context->txn_->get_state() == TransactionState::ABORTED) {
            context->txn_ = txn_manager_->begin(nullptr, context->log_mgr_);
            *txn_id = context->txn_->get_transaction_id();
            context->txn_->set_txn_mode(false);
        }
    }

    void read_data() {
        const char* select_small = "select";
        const char* select_big = "SELECT";
        char data_recv[BUFFER_LENGTH];
        std::string data_send("");

        while (true) {
            {
                std::unique_lock<std::mutex> lock(mtx_);
                cv_.wait(lock, [this] {
                    if (sql_queue_.size() >= 1000) {
                        ++read_wait_;
                    }
                    return sql_queue_.size() < 1000;
                });
                lock.unlock();
            }

            if (exit_) {
                break;
            }

            memset(data_recv, 0, BUFFER_LENGTH);

            int rbytes = read(fd_, data_recv, BUFFER_LENGTH);
            if (rbytes == 0) {
                exit_ = true;
                std::cout << "Maybe the client has closed, wait: " << read_wait_ << std::endl;
                break;
            } else if (rbytes == -1) {
                exit_ = true;
                std::cout << "Client read error!" << std::endl;
                break;
            }

            if (strcmp(data_recv, "exit") == 0) {
                exit_ = true;
                std::cout << "Client exit." << std::endl;
                break;
            }
            if (strcmp(data_recv, "crash") == 0) {
                exit_ = true;
                std::cout << "Server crash" << std::endl;
                break;
            }
            if (strcmp(data_recv, "set output_file off") == 0) {
                ql_manager_->write_output_ = false;
                write(fd_, data_send.c_str(), 1);
                continue;
            }

            ++read_;
            if (read_ > 100) {
                write(fd_, data_send.c_str(), 1);
            }

            std::unique_lock<std::mutex> lock(mtx_);
            sql_queue_.push(data_recv);
            cv_.notify_all();
        }
        cv_.notify_all();
    }

    void process_sql() {
        const char* select_small = "select";
        const char* select_big = "SELECT";
        char data_send[BUFFER_LENGTH];
        int offset = 0;
        txn_id_t txn_id = INVALID_TXN_ID;

        while (true) {
            std::unique_lock<std::mutex> lock(mtx_);
            cv_.wait(lock, [this] {
                if (sql_queue_.empty() && !exit_) {
                    ++sql_wait_;
                }
                return !sql_queue_.empty() || exit_;
            });

            if (exit_ && sql_queue_.empty()) {
                std::cout << "process thread exit, wait: " << sql_wait_ << std::endl;
                break;
            }

            std::string sql = sql_queue_.front();
            sql_queue_.pop();
            lock.unlock();

            memset(data_send, '\0', BUFFER_LENGTH);
            offset = 0;

            // 开启事务，初始化系统所需的上下文信息（包括事务对象指针、锁管理器指针、日志管理器指针、存放结果的buffer、记录结果长度的变量）
            Context* context = new Context(lock_manager_, log_manager_, nullptr, data_send, &offset);
            set_transaction(&txn_id, context);

            // 用于判断是否已经调用了yy_delete_buffer来删除buf
            bool finish_analyze = false;
            pthread_mutex_lock(buffer_mutex_);
            YY_BUFFER_STATE buf = yy_scan_string(sql.c_str());
            if (yyparse() == 0) {
                if (ast::parse_tree != nullptr) {
                    try {
                        // analyze and rewrite
                        std::shared_ptr<Query> query = analyze_->do_analyze(ast::parse_tree);
                        yy_delete_buffer(buf);
                        finish_analyze = true;
                        pthread_mutex_unlock(buffer_mutex_);
                        // 优化器
                        std::shared_ptr<Plan> plan = optimizer_->plan_query(cache, query, context);
                        // portal
                        std::shared_ptr<PortalStmt> portalStmt = portal_->start(plan, context, cache);
                        portal_->run(portalStmt, ql_manager_, &txn_id, context);
                        portal_->drop();
                    } catch (TransactionAbortException& e) {
                        // 事务需要回滚，需要把abort信息返回给客户端并写入output.txt文件中
                        std::string str = "abort\n";
                        memcpy(data_send, str.c_str(), str.length());
                        data_send[str.length()] = '\0';
                        offset = str.length();

                        // 回滚事务
                        txn_manager_->abort(context->txn_, log_manager_);
                        std::cout << e.GetInfo() << std::endl;

                        std::fstream outfile;
                        outfile.open("output.txt", std::ios::out | std::ios::app);
                        outfile << str;
                        outfile.close();
                    } catch (RMDBError& e) {
                        // 遇到异常，需要打印failure到output.txt文件中，并发异常信息返回给客户端
                        std::cerr << e.what() << std::endl;

                        memcpy(data_send, e.what(), e.get_msg_len());
                        data_send[e.get_msg_len()] = '\n';
                        data_send[e.get_msg_len() + 1] = '\0';
                        offset = e.get_msg_len() + 1;

                        // 将报错信息写入output.txt
                        std::fstream outfile;
                        outfile.open("output.txt", std::ios::out | std::ios::app);
                        outfile << "failure\n";
                        outfile.close();
                    }
                }
            } else {
                std::fstream outfile;
                outfile.open("output.txt", std::ios::out | std::ios::app);
                outfile << "failure\n";
                outfile.close();
            }
            if (finish_analyze == false) {
                yy_delete_buffer(buf);
                pthread_mutex_unlock(buffer_mutex_);
            }
            // future TODO: 格式化 sql_handler.result, 传给客户端
            // send result with fixed format, use protobuf in the future
            ++write_;
            if (write_ <= 100) {
                if (write(fd_, data_send, offset + 1) == -1) {
                    delete context->txn_;
                    context->txn_ = nullptr;
                    delete context;
                    exit_ = true;
                    break;
                }
            }
            // 如果是单挑语句，需要按照一个完整的事务来执行，所以执行完当前语句后，自动提交事务
            if (context->txn_ && context->txn_->get_txn_mode() == false) {
                txn_manager_->commit(context->txn_, context->log_mgr_);
                delete context->txn_;
                context->txn_ = nullptr;
            }
            delete context;
            cv_.notify_all();
        }
    }

private:
    int fd_;
    std::mutex mtx_{};
    std::condition_variable cv_{};
    std::queue<std::string> sql_queue_{};
    std::atomic<bool> exit_;
    std::thread read_thread_;
    std::thread sql_thread_;

    LockManager* lock_manager_;
    TransactionManager* txn_manager_;
    Optimizer* optimizer_;
    QlManager* ql_manager_;
    LogManager* log_manager_;
    Portal* portal_;
    Analyze* analyze_;

    pthread_mutex_t* buffer_mutex_;

    int read_wait_;
    int sql_wait_;
    int read_;
    int write_;

    Aggregate_cache* cache;
};
