/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cstring>
#include <memory>
#include <string>
#include <vector>

#include "parser/ast.h"
#include "parser/parser.h"

// FIXME: Too many code
typedef enum PlanTag {
    T_Invalid = 1,
    T_Help,
    T_ShowTable,
    T_DescTable,
    T_ShowIndex,
    T_CreateTable,
    T_DropTable,
    T_CreateIndex,
    T_DropIndex,
    T_SetKnob,
    T_Insert,
    T_Update,
    T_Delete,
    T_Load,
    T_select,
    T_Transaction_begin,
    T_Transaction_commit,
    T_Transaction_abort,
    T_Transaction_rollback,
    T_SeqScan,
    T_IndexScan,
    T_NestLoop,
    T_SortMerge,  // sort merge join
    T_Sort,
    T_Predicate,
    T_Aggregate,
    T_Groupby,
    T_Projection,
    T_Print
} PlanTag;

class PlanTagHelper {
public:
    static std::string ToString(PlanTag tag) {
        switch (tag) {
            case T_Invalid:
                return "Invalid";
            case T_Help:
                return "Help";
            case T_ShowTable:
                return "ShowTable";
            case T_DescTable:
                return "DescTable";
            case T_ShowIndex:
                return "ShowIndex";
            case T_CreateTable:
                return "CreateTable";
            case T_DropTable:
                return "DropTable";
            case T_CreateIndex:
                return "CreateIndex";
            case T_DropIndex:
                return "DropIndex";
            case T_SetKnob:
                return "SetKnob";
            case T_Insert:
                return "Insert";
            case T_Update:
                return "Update";
            case T_Delete:
                return "Delete";
            case T_Load:
                return "Load";
            case T_select:
                return "Select";
            case T_Transaction_begin:
                return "Transaction_begin";
            case T_Transaction_commit:
                return "Transaction_commit";
            case T_Transaction_abort:
                return "Transaction_abort";
            case T_Transaction_rollback:
                return "Transaction_rollback";
            case T_SeqScan:
                return "SeqScan";
            case T_IndexScan:
                return "IndexScan";
            case T_NestLoop:
                return "NestLoop";
            case T_SortMerge:
                return "SortMerge";
            case T_Sort:
                return "Sort";
            case T_Predicate:
                return "Predicate";
            case T_Aggregate:
                return "Aggregate";
            case T_Groupby:
                return "Groupby";
            case T_Projection:
                return "Projection";
            case T_Print:
                return "Print";
            default:
                return "Unknown";
        }
    }
};

// 查询执行计划
class Plan {
public:
    PlanTag tag;
    virtual ~Plan() = default;
};

class ScanPlan : public Plan {
public:
    ScanPlan(PlanTag tag, SmManager* sm_manager, std::string tab_name, std::vector<std::string> index_col_names,
             std::vector<Condition> conds, int index_cond_num = 0) {
        Plan::tag = tag;
        tab_name_ = std::move(tab_name);
        TabMeta& tab = sm_manager->db_.get_table(tab_name_);
        cols_ = tab.cols;
        index_col_names_ = index_col_names;
        len_ = cols_.back().offset + cols_.back().len;
        conds_ = conds;
        index_cond_num_ = index_cond_num;
    }
    ~ScanPlan() {}

public:
    std::string tab_name_;
    std::vector<ColMeta> cols_;
    std::vector<std::string> index_col_names_;
    size_t len_;
    std::vector<Condition> conds_;
    int index_cond_num_;
};

class JoinPlan : public Plan {
public:
    JoinPlan(PlanTag tag, std::shared_ptr<Plan> left, std::shared_ptr<Plan> right, std::vector<Condition> conds) {
        Plan::tag = tag;
        left_ = std::move(left);
        right_ = std::move(right);
        conds_ = std::move(conds);
        type = INNER_JOIN;
    }
    ~JoinPlan() {}

public:
    // 左节点
    std::shared_ptr<Plan> left_;
    // 右节点
    std::shared_ptr<Plan> right_;
    // 连接条件
    std::vector<Condition> conds_;
    // future TODO: 后续可以支持的连接类型
    JoinType type;
};

class PredicatePlan : public Plan {
public:
    PredicatePlan(PlanTag tag, std::shared_ptr<Plan> scan_plan, std::vector<std::shared_ptr<Plan>>&& sub_queries,
                  std::vector<Condition> conds, std::string tab_name, int index_cond_num) {
        Plan::tag = tag;
        scan_plan_ = std::move(scan_plan);
        sub_queries_ = std::move(sub_queries);
        conds_ = std::move(conds);
        tab_name_ = std::move(tab_name);
        index_cond_num_ = index_cond_num;
    }

public:
    std::shared_ptr<Plan> scan_plan_{};
    std::vector<std::shared_ptr<Plan>> sub_queries_{};
    std::vector<Condition> conds_{};
    std::string tab_name_;
    int index_cond_num_{};
};

class AggregatePlan : public Plan {
public:
    AggregatePlan(PlanTag tag, std::shared_ptr<Plan> child, std::vector<TabCol> aggr_cols) {
        Plan::tag = tag;
        subplan_ = std::move(child);
        aggr_cols_ = std::move(aggr_cols);
    }

public:
    std::shared_ptr<Plan> subplan_{};
    std::vector<TabCol> aggr_cols_{};
};

class GroupbyPlan : public Plan {
public:
    GroupbyPlan(PlanTag tag, std::shared_ptr<Plan> child, std::vector<TabCol> group_cols,
                std::vector<Condition> having_conds) {
        Plan::tag = tag;
        subplan_ = std::move(child);
        group_cols_ = std::move(group_cols);
        having_conds_ = std::move(having_conds);
    }

public:
    std::shared_ptr<Plan> subplan_{};
    std::vector<TabCol> group_cols_{};
    std::vector<Condition> having_conds_{};
};

class ProjectionPlan : public Plan {
public:
    ProjectionPlan(PlanTag tag, std::shared_ptr<Plan> subplan, std::vector<TabCol> sel_cols) {
        Plan::tag = tag;
        subplan_ = std::move(subplan);
        sel_cols_ = std::move(sel_cols);
    }
    ~ProjectionPlan() {}

public:
    std::shared_ptr<Plan> subplan_;
    std::vector<TabCol> sel_cols_;
};

class SortPlan : public Plan {
public:
    SortPlan(PlanTag tag, std::shared_ptr<Plan> subplan, std::vector<TabCol> order_cols, std::vector<bool> is_descs,
             int limit_cnt) {
        Plan::tag = tag;
        subplan_ = std::move(subplan);
        order_cols_ = std::move(order_cols);
        is_descs_ = is_descs;
        limit_cnt_ = limit_cnt;
    }
    ~SortPlan() {}

public:
    std::shared_ptr<Plan> subplan_;
    std::vector<TabCol> order_cols_;
    std::vector<bool> is_descs_;
    int limit_cnt_;
};

class PrintPlan : public Plan {
public:
    PrintPlan(PlanTag tag, std::shared_ptr<Plan> subplan, std::string file_name) {
        Plan::tag = tag;
        subplan_ = std::move(subplan);
    }
    ~PrintPlan() {}

public:
    std::shared_ptr<Plan> subplan_{};
    std::string file_name_{};
};

class LoadPlan : public Plan {
public:
    LoadPlan(PlanTag tag, std::string tab_name, std::string file_name) {
        Plan::tag = tag;
        tab_name_ = std::move(tab_name);
        file_name_ = std::move(file_name);
    }
    ~LoadPlan() {}

public:
    std::string tab_name_;
    std::string file_name_;
};

// dml语句，包括insert; delete; update; select语句　
class DMLPlan : public Plan {
public:
    DMLPlan(PlanTag tag, std::shared_ptr<Plan> subplan, std::string tab_name, std::vector<Value> values,
            std::vector<Condition> conds, std::vector<SetClause> set_clauses) {
        Plan::tag = tag;
        subplan_ = std::move(subplan);
        tab_name_ = std::move(tab_name);
        values_ = std::move(values);
        conds_ = std::move(conds);
        set_clauses_ = std::move(set_clauses);
    }
    ~DMLPlan() {}

public:
    std::shared_ptr<Plan> subplan_;
    std::string tab_name_;
    std::vector<Value> values_;
    std::vector<Condition> conds_;
    std::vector<SetClause> set_clauses_;
};

// ddl语句, 包括create/drop table; create/drop index;
class DDLPlan : public Plan {
public:
    DDLPlan(PlanTag tag, std::string tab_name, std::vector<std::string> col_names, std::vector<ColDef> cols) {
        Plan::tag = tag;
        tab_name_ = std::move(tab_name);
        cols_ = std::move(cols);
        tab_col_names_ = std::move(col_names);
    }
    ~DDLPlan() {}

public:
    std::string tab_name_;
    std::vector<std::string> tab_col_names_;
    std::vector<ColDef> cols_;
};

// help; show tables; desc tables; begin; abort; commit; rollback语句对应的plan
class OtherPlan : public Plan {
public:
    OtherPlan(PlanTag tag, std::string tab_name) {
        Plan::tag = tag;
        tab_name_ = std::move(tab_name);
    }
    ~OtherPlan() {}

public:
    std::string tab_name_;
};

// Set Knob Plan
class SetKnobPlan : public Plan {
public:
    SetKnobPlan(ast::SetKnobType knob_type, bool bool_value) {
        Plan::tag = T_SetKnob;
        set_knob_type_ = knob_type;
        bool_value_ = bool_value;
    }

public:
    ast::SetKnobType set_knob_type_;
    bool bool_value_;
};

class plannerInfo {
public:
    std::shared_ptr<ast::SelectStmt> parse;
    std::vector<Condition> where_conds;
    std::vector<TabCol> sel_cols;
    std::shared_ptr<Plan> plan;
    std::vector<std::shared_ptr<Plan>> table_scan_executors;
    std::vector<SetClause> set_clauses;
    plannerInfo(std::shared_ptr<ast::SelectStmt> parse_) : parse(std::move(parse_)) {}
};
