/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "planner.h"

#include <memory>

#include "execution/executor_delete.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_insert.h"
#include "execution/executor_nestedloop_join.h"
#include "execution/executor_print.h"
#include "execution/executor_projection.h"
#include "execution/executor_seq_scan.h"
#include "execution/executor_update.h"
#include "index/ix.h"
#include "optimizer/plan.h"
// #include "optimizer/aggregate_cache.h"
#include "record_printer.h"
#include "system/sm_meta.h"

// 目前的索引匹配规则为：完全匹配索引字段，且全部为单点查询，不会自动调整where条件的顺序
// 调整条件顺序，索引条件在前，其他条件在后
// 返回值是索引条件的个数，不包括 >= > = 这类下限索引条件
int Planner::get_index_cols(std::string tab_name, std::vector<Condition>& curr_conds,
                            std::vector<std::string>& index_col_names) {
    std::unordered_map<std::string, std::vector<Condition>> col_conds_map;
    std::unordered_set<std::string> curr_conds_set;  // 处理插入过的列名
    std::vector<Condition> curr_conds_copy;
    TabMeta& tab = sm_manager_->db_.get_table(tab_name);
    int index_cond = -1;

    index_col_names.clear();
    for (auto& cond : curr_conds) {
        if (col_conds_map.count(cond.lhs_col.col_name) == 0) {
            col_conds_map.emplace(cond.lhs_col.col_name, std::vector<Condition>());
            if (cond.rhs_type == !RhsType::COLUMN && cond.lhs_col.tab_name.compare(tab_name) == 0) {
                index_col_names.push_back(cond.lhs_col.col_name);
            }
        }
        col_conds_map[cond.lhs_col.col_name].push_back(cond);
    }
    // 无条件查询，即没有where
    if (curr_conds.empty()) {
        for (auto& col_meta : tab.cols) {
            index_col_names.push_back(col_meta.name);
        }
    }
    // 处理条件顺序按照索引顺序
    if (tab.is_index(index_col_names)) {
        if (curr_conds.empty()) {
            return 0;
        }
        for (auto& col_name : index_col_names) {
            if (col_conds_map.find(col_name) == col_conds_map.end()) {
                continue;
            }
            std::copy(col_conds_map[col_name].begin(), col_conds_map[col_name].end(),
                      std::back_inserter(curr_conds_copy));
            curr_conds_set.emplace(col_name, 0);
        }
        index_cond = curr_conds_copy.size();
        for (auto& cond : curr_conds) {
            if (curr_conds_set.find(cond.lhs_col.col_name) == curr_conds_set.end()) {
                curr_conds_copy.push_back(cond);
            }
        }
        curr_conds.clear();
        std::copy(curr_conds_copy.begin(), curr_conds_copy.end(), std::back_inserter(curr_conds));
    }
    return index_cond;
}

bool Planner::have_index(std::string tab_name, std::vector<TabCol>& cols, std::vector<std::string>& index_col_names) {
    TabMeta& tab = sm_manager_->db_.get_table(tab_name);
    bool have_index = false;

    for (auto& index : tab.indexes) {
        if (index.col_num == cols.size()) {
            size_t cnt = 0;
            for (auto& col : cols) {
                for (auto& index_col : index.cols) {
                    if (col.col_name.compare(index_col.name) == 0) {
                        cnt++;
                        break;
                    }
                }
            }
            if (cnt == cols.size()) {
                have_index = true;
                for (auto& col_meta : index.cols) {
                    index_col_names.push_back(col_meta.name);
                }
                break;
            }
        }
    }

    return have_index;
}

/**
 * @brief 表算子条件谓词生成
 *
 * @param conds 条件
 * @param tab_names 表名
 * @return std::vector<Condition>
 */
// left cond must be column type
std::vector<Condition> pop_conds(std::vector<Condition>& conds, std::string tab_name) {
    std::vector<Condition> solved_conds;
    auto it = conds.begin();
    while (it != conds.end()) {
        if (tab_name.compare(it->lhs_col.tab_name) != 0) {
            ++it;
        } else {
            if (it->rhs_type != RhsType::COLUMN) {
                solved_conds.emplace_back(std::move(*it));
                it = conds.erase(it);
            } else {
                if (tab_name.compare(it->rhs_col.tab_name) == 0) {
                    solved_conds.emplace_back(std::move(*it));
                    it = conds.erase(it);
                } else {
                    ++it;
                }
            }
        }
    }
    return solved_conds;
}

void RewriteConds(std::vector<Condition>& conds, std::string& tab_name) {
    std::map<std::string, std::vector<Condition>> cond_map;
    std::map<std::string, bool> index_map;
    for (auto& cond : conds) {
        auto& col_name = cond.lhs_col.col_name;
        if (cond_map.find(col_name) == cond_map.end()) {
            cond_map[col_name] = std::vector<Condition>();
        }
        cond_map[col_name].push_back(cond);
        index_map[col_name] = true;
    }
    for (auto& [col_name, conds] : cond_map) {
        if (conds.size() == 1) {
            continue;
        }
        Value min_val;
        Value max_val;
        Value equal_val;
        CompOp min_op = OP_GT;
        CompOp max_op = OP_LT;
        int min_idx = -1;
        int max_idx = -1;
        int equal_idx = -1;
        int idx = 0;
        for (auto& cond : conds) {
            if (cond.rhs_type != RhsType::VALUE) {
                ++idx;
                continue;
            }
            if (cond.op == OP_EQ) {
                if (equal_idx != -1) {
                    index_map[col_name] = false;
                    break;
                }
                equal_val = cond.rhs_val;
                equal_idx = idx;
            } else if (cond.op == OP_GE || cond.op == OP_GT) {
                if (min_idx == -1) {
                    min_val = cond.rhs_val;
                    min_op = cond.op;
                    min_idx = idx;
                } else if (min_val < cond.rhs_val) {
                    min_val = cond.rhs_val;
                    min_op = cond.op;
                    min_idx = idx;
                } else if (min_val == cond.rhs_val && min_op == OP_GE && cond.op == OP_GT) {
                    min_op = OP_GT;
                }
            } else if (cond.op == OP_LE || cond.op == OP_LT) {
                if (max_idx == -1) {
                    max_val = cond.rhs_val;
                    max_op = cond.op;
                    max_idx = idx;
                } else if (max_val > cond.rhs_val) {
                    max_val = cond.rhs_val;
                    max_op = cond.op;
                    max_idx = idx;
                } else if (max_val == cond.rhs_val && max_op == OP_LE && cond.op == OP_LT) {
                    max_op = OP_LT;
                }
            }
            ++idx;
        }
        if (!index_map[col_name]) {
            continue;
        }
        if (min_idx != -1 && max_idx != -1) {
            int cmp = min_val.Compare(max_val);
            if (cmp > 0) {
                index_map[col_name] = false;
            } else if (cmp == 0) {
                if (min_op == OP_GT || max_op == OP_LT) {
                    index_map[col_name] = false;
                } else {
                    if (equal_idx != -1) {
                        if (equal_val.Compare(min_val) != 0) {
                            index_map[col_name] = false;
                        } else {
                            Condition cond = conds[equal_idx];
                            conds.clear();
                            conds.push_back(cond);
                        }
                    } else {
                        Condition cond;
                        cond.lhs_col = conds[min_idx].lhs_col;
                        cond.op = OP_EQ;
                        cond.rhs_type = RhsType::VALUE;
                        cond.rhs_val = min_val;
                        conds.clear();
                        conds.push_back(cond);
                    }
                }
            } else {
                if (equal_idx != -1) {
                    int min_cmp = equal_val.Compare(min_val);
                    int max_cmp = equal_val.Compare(max_val);
                    if (min_cmp < 0 || (min_cmp == 0 && min_op == OP_GT) || max_cmp > 0 ||
                        (max_cmp == 0 && max_op == OP_LT)) {
                        index_map[col_name] = false;
                    } else {
                        Condition cond = conds[equal_idx];
                        conds.clear();
                        conds.push_back(cond);
                    }
                } else {
                    Condition min_cond = conds[min_idx];
                    Condition max_cond = conds[max_idx];
                    conds.clear();
                    conds.push_back(min_cond);
                    conds.push_back(max_cond);
                }
            }
        } else if (min_idx != -1) {
            if (equal_idx != -1) {
                int cmp = equal_val.Compare(min_val);
                if (cmp < 0 || (cmp == 0 && min_op == OP_GT)) {
                    index_map[col_name] = false;
                } else {
                    Condition cond = conds[equal_idx];
                    conds.clear();
                    conds.push_back(cond);
                }
            } else {
                Condition cond = conds[min_idx];
                conds.clear();
                conds.push_back(cond);
            }
        } else if (max_idx != -1) {
            if (equal_idx != -1) {
                int cmp = equal_val.Compare(max_val);
                if (cmp > 0 || (cmp == 0 && max_op == OP_LT)) {
                    index_map[col_name] = false;
                } else {
                    Condition cond = conds[equal_idx];
                    conds.clear();
                    conds.push_back(cond);
                }
            } else {
                Condition cond = conds[max_idx];
                conds.clear();
                conds.push_back(cond);
            }
        }
    }

    conds.clear();
    for (auto& [col_name, ok] : index_map) {
        if (ok) {
            std::copy(cond_map[col_name].begin(), cond_map[col_name].end(), std::back_inserter(conds));
        }
    }
}

std::shared_ptr<Query> Planner::logical_optimization(std::shared_ptr<Query> query, Context* context) {

    // TODO 实现逻辑优化规则

    return query;
}

std::shared_ptr<Plan> Planner::physical_optimization(Aggregate_cache *aggregation_cache, std::shared_ptr<Query> query, Context* context) {
    std::shared_ptr<Plan> plan = make_plan(aggregation_cache, query, context);

    // 其他物理优化
    return plan;
}

std::shared_ptr<Plan> Planner::make_plan(Aggregate_cache *aggregation_cache, std::shared_ptr<Query> query, Context* context) {
    auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse);
    std::vector<std::string> tables = query->tables;
    std::vector<std::shared_ptr<Plan>> table_scan_executors(tables.size());

    // Here has removed all single table conditions
    for (size_t i = 0; i < tables.size(); i++) {
        std::shared_ptr<Plan> table_scan_executor = nullptr;
        auto curr_conds = pop_conds(query->where_conds, tables[i]);
        // tidy conditions with single table
        RewriteConds(curr_conds, tables[i]);

        // separate conditions by whether has subquery
        std::vector<Condition> scan_conds;
        std::vector<Condition> subquery_conds;
        for (auto& cond : curr_conds) {
            if (cond.rhs_type == RhsType::SUB_QUERY || cond.rhs_type == RhsType::COLUMN) {
                subquery_conds.push_back(cond);
            } else {
                scan_conds.push_back(cond);
            }
        }

        std::vector<std::string> index_col_names;
        int index_cond = get_index_cols(tables[i], curr_conds, index_col_names);
        // if seqscan, we use pass scan_conds to RmScan, otherwise we pass curr_conds
        if (index_cond == -1) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executor =
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, tables[i], index_col_names, scan_conds);
        } else {  // 存在索引
            table_scan_executor = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, tables[i], index_col_names,
                                                             curr_conds, index_cond);
        }

        std::vector<std::shared_ptr<Plan>> sub_queries;
        for (auto& cond : subquery_conds) {
            if (cond.rhs_type == RhsType::SUB_QUERY) {
                sub_queries.emplace_back(generate_select_plan(aggregation_cache, cond.rhs_sub_query, nullptr));
            }
        }

        if (table_scan_executor->tag == T_SeqScan) {
            table_scan_executors[i] = !subquery_conds.empty()
                ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor), std::move(sub_queries),
                                                  subquery_conds, tables[i], index_cond)
                : std::move(table_scan_executor);
        } else {
            table_scan_executors[i] = !curr_conds.empty()
                ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor), std::move(sub_queries),
                                                  curr_conds, tables[i], index_cond)
                : std::move(table_scan_executor);
        }
    }

    std::shared_ptr<Plan> left = table_scan_executors[0];
    std::shared_ptr<Plan> right = nullptr;
    std::shared_ptr<Plan> root = left;
    for (size_t i = 1; i < tables.size(); ++i) {
        right = left;
        left = table_scan_executors[i];
        std::vector<Condition> join_conds;
        if (x->has_sort) {
            // if (x->has_sort && enable_sortmerge_join) {
            root = std::make_shared<JoinPlan>(T_SortMerge, left, right, join_conds);
        } else {
            root = std::make_shared<JoinPlan>(T_NestLoop, left, right, join_conds);
        }
    }

    // condition push down
    auto conds = std::move(query->where_conds);
    std::shared_ptr<Plan> node = root;
    std::shared_ptr<JoinPlan> y = nullptr;
    std::map<CompOp, CompOp> swap_op = {
        {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
    };
    while ((y = std::dynamic_pointer_cast<JoinPlan>(node))) {
        auto left = std::dynamic_pointer_cast<ScanPlan>(y->left_);
        auto right = std::dynamic_pointer_cast<JoinPlan>(y->right_);
        if (!left) {
            auto filter = std::dynamic_pointer_cast<PredicatePlan>(y->left_);
            left = std::dynamic_pointer_cast<ScanPlan>(filter->scan_plan_);
        }
        auto it = conds.begin();
        while (it != conds.end()) {
            if (it->lhs_col.tab_name.compare(left->tab_name_) == 0) {
                y->conds_.emplace_back(std::move(*it));
                it = conds.erase(it);
            } else if (it->rhs_col.tab_name.compare(left->tab_name_) == 0) {
                std::swap(it->lhs_col, it->rhs_col);
                it->op = swap_op.at(it->op);
                y->conds_.emplace_back(std::move(*it));
                it = conds.erase(it);
            } else {
                it++;
            }
        }
        node = y->right_;
    }

    // handle sortmerge join, add a sort operator on the table scan
    node = root;
    bool is_join = false;
    std::shared_ptr<JoinPlan> prev_y = nullptr;
    while (x->has_sort && (y = std::dynamic_pointer_cast<JoinPlan>(node))) {
        // while (x->has_sort && enable_sortmerge_join && (y = std::dynamic_pointer_cast<JoinPlan>(node))) {
        auto left = std::dynamic_pointer_cast<ScanPlan>(y->left_);
        auto right = std::dynamic_pointer_cast<JoinPlan>(y->right_);
        auto filter = std::dynamic_pointer_cast<PredicatePlan>(y->left_);
        if (!left) {
            left = std::dynamic_pointer_cast<ScanPlan>(filter->scan_plan_);
        }
        std::vector<TabCol> order_cols;
        std::vector<bool> is_descs;
        for (auto& cond : y->conds_) {
            if (cond.lhs_col.tab_name.compare(left->tab_name_) == 0) {
                order_cols.emplace_back(cond.lhs_col);
                is_descs.push_back(false);
            }
        }
        // use index scan replace seq scan
        std::vector<std::string> index_col_names;
        if (left->tag == T_SeqScan && have_index(left->tab_name_, order_cols, index_col_names)) {
            std::vector<Condition> conds = left->conds_;
            left = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, left->tab_name_, index_col_names, conds, 0);
            std::shared_ptr<PrintPlan> print =
                std::make_shared<PrintPlan>(T_Print, std::move(left), std::string("sorted_results.txt"));
            if (!filter) {
                y->left_ = print;
            } else {
                filter->scan_plan_ = print;
            }
        } else {
            std::shared_ptr<SortPlan> sort = std::make_shared<SortPlan>(T_Sort, left, order_cols, is_descs, 0);
            std::shared_ptr<PrintPlan> print =
                std::make_shared<PrintPlan>(T_Print, std::move(sort), std::string("sorted_results.txt"));
            if (!left) {
                filter->scan_plan_ = std::move(print);
            } else {
                y->left_ = std::move(print);
            }
        }

        node = y->right_;
        prev_y = y;
        is_join = true;
    }
    if (is_join) {
        std::shared_ptr<PredicatePlan> temp;
        std::shared_ptr<ScanPlan> z;
        if ((temp = std::dynamic_pointer_cast<PredicatePlan>(node))) {
            z = std::dynamic_pointer_cast<ScanPlan>(temp->scan_plan_);
        } else {
            z = std::dynamic_pointer_cast<ScanPlan>(node);
        }
        std::vector<TabCol> order_cols;
        std::vector<bool> is_descs;
        for (auto& cond : prev_y->conds_) {
            if (cond.rhs_col.tab_name.compare(z->tab_name_) == 0) {
                order_cols.emplace_back(cond.rhs_col);
                is_descs.push_back(false);
            }
        }
        // use index scan replace seq scan
        std::vector<std::string> index_col_names;
        if (z->tag == T_SeqScan && have_index(z->tab_name_, order_cols, index_col_names)) {
            std::vector<Condition> conds = z->conds_;
            z = std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, z->tab_name_, index_col_names, conds, 0);
            std::shared_ptr<PrintPlan> print =
                std::make_shared<PrintPlan>(T_Print, std::move(z), std::string("sorted_results.txt"));
            if (temp) {
                temp->scan_plan_ = std::move(print);
            } else {
                prev_y->right_ = std::move(print);
            }
        } else {
            std::shared_ptr<SortPlan> sort = std::make_shared<SortPlan>(T_Sort, z, order_cols, is_descs, 0);
            std::shared_ptr<PrintPlan> print =
                std::make_shared<PrintPlan>(T_Print, std::move(sort), std::string("sorted_results.txt"));
            if (temp) {
                temp->scan_plan_ = std::move(print);
            } else {
                prev_y->right_ = std::move(print);
            }
        }
    }

    // groupby, and add a projection before
    std::vector<TabCol> project_cols = query->cols;
    for (auto& col : query->groupby_cols) {
        if (std::find(project_cols.begin(), project_cols.end(), col) == project_cols.end()) {
            project_cols.push_back(col);
        }
    }
    for (auto& cond : query->having_conds) {
        if (std::find(project_cols.begin(), project_cols.end(), cond.lhs_col) == project_cols.end()) {
            project_cols.push_back(cond.lhs_col);
        }
        if (cond.rhs_type == RhsType::COLUMN &&
            std::find(project_cols.begin(), project_cols.end(), cond.rhs_col) == project_cols.end()) {
            project_cols.push_back(cond.rhs_col);
        }
    }

    std::shared_ptr<Plan> projection = std::make_shared<ProjectionPlan>(T_Projection, std::move(root), project_cols);
    std::shared_ptr<Plan> groupby = !query->groupby_cols.empty()
        ? std::make_shared<GroupbyPlan>(T_Groupby, std::move(projection), query->groupby_cols, query->having_conds)
        : std::move(projection);

    // aggregate
    bool has_aggregate = false;
    for (auto& col : query->cols) {
        if (col.aggr_op != AggrOp::AGGR_NONE) {
            has_aggregate = true;
            break;
        }
    }

    //Aggregate cache, conditions
    if (has_aggregate == true) {
        aggregation_cache->add_cache();
        aggregation_cache->get_cache()[aggregation_cache->get_location()].table_list = query->tables;
        (aggregation_cache->get_cache())[aggregation_cache->get_location()].where_conds = query->where_conds;
        (aggregation_cache->get_cache())[aggregation_cache->get_location()].having_conds = query->having_conds;
        (aggregation_cache->get_cache())[aggregation_cache->get_location()].groupby_cols = query->groupby_cols;
        for (auto& col : query->cols) {
            if (col.aggr_op == AggrOp::AGGR_COUNT) (aggregation_cache->get_cache())[aggregation_cache->get_location()].count_col.push_back(col);
            else if (col.aggr_op == AggrOp::AGGR_MAX) (aggregation_cache->get_cache())[aggregation_cache->get_location()].max_col.push_back(col);
            else if (col.aggr_op == AggrOp::AGGR_MIN) (aggregation_cache->get_cache())[aggregation_cache->get_location()].min_col.push_back(col);
            else if (col.aggr_op == AggrOp::AGGR_SUM) (aggregation_cache->get_cache())[aggregation_cache->get_location()].sum_col.push_back(col);
        }
    }

    std::shared_ptr<Plan> aggregate = !query->groupby_cols.empty() || has_aggregate
        ? std::make_shared<AggregatePlan>(T_Aggregate, std::move(groupby), query->cols)
        : std::move(groupby);

    // return aggregate;

    ///
    // orderby
    // if no orderby, or has handled in sortmerge join, exit directly
    if (!x->has_sort || (is_join)) {
        // if (!x->has_sort || (enable_sortmerge_join && is_join)) {
        return aggregate;
    }

    std::vector<bool> is_descs;
    for (auto& col : x->order->cols) {
        is_descs.push_back(col->order_by_dir == ast::OrderBy_DESC);
    }
    // TODO: join的排序与最后排序要求不同
    std::shared_ptr<Plan> orderby =
        std::make_shared<SortPlan>(T_Sort, std::move(aggregate), query->order_cols, is_descs, x->order->limit_cnt);

    return orderby;
}

/**
 * @brief select plan 生成
 *
 * @param sel_cols select plan 选取的列
 * @param tab_names select plan 目标的表
 * @param conds select plan 选取条件
 */
std::shared_ptr<Plan> Planner::generate_select_plan(Aggregate_cache* aggregation_cache, std::shared_ptr<Query> query, Context* context) {
    // 逻辑优化
    query = logical_optimization(std::move(query), context);

    // 物理优化
    auto sel_cols = query->cols;
    std::shared_ptr<Plan> plannerRoot = physical_optimization(aggregation_cache, query, context);
    plannerRoot = std::make_shared<ProjectionPlan>(T_Projection, std::move(plannerRoot), std::move(sel_cols));

    return plannerRoot;
}

// 生成DDL语句和DML语句的查询执行计划
std::shared_ptr<Plan> Planner::do_planner(Aggregate_cache *aggregation_cache, std::shared_ptr<Query> query, Context* context) {
    std::shared_ptr<Plan> plannerRoot;
    if (auto x = std::dynamic_pointer_cast<ast::CreateTable>(query->parse)) {
        // create table;
        std::vector<ColDef> col_defs;
        for (auto& field : x->fields) {
            if (auto sv_col_def = std::dynamic_pointer_cast<ast::ColDef>(field)) {
                ColDef col_def = {.name = sv_col_def->col_name,
                                  .type = interp_sv_type(sv_col_def->type_len->type),
                                  .len = sv_col_def->type_len->len};
                col_defs.push_back(col_def);
            } else {
                throw InternalError("Unexpected field type");
            }
        }
        plannerRoot = std::make_shared<DDLPlan>(T_CreateTable, x->tab_name, std::vector<std::string>(), col_defs);
    } else if (auto x = std::dynamic_pointer_cast<ast::DropTable>(query->parse)) {
        // drop table;
        plannerRoot =
            std::make_shared<DDLPlan>(T_DropTable, x->tab_name, std::vector<std::string>(), std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::CreateIndex>(query->parse)) {
        // create index;
        plannerRoot = std::make_shared<DDLPlan>(T_CreateIndex, x->tab_name, x->col_names, std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::DropIndex>(query->parse)) {
        // drop index
        plannerRoot = std::make_shared<DDLPlan>(T_DropIndex, x->tab_name, x->col_names, std::vector<ColDef>());
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(query->parse)) {
        // insert;
        plannerRoot = std::make_shared<DMLPlan>(T_Insert, std::shared_ptr<Plan>(), x->tab_name, query->values,
                                                std::vector<Condition>(), std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(query->parse)) {
        // delete;
        // 生成表扫描方式
        std::shared_ptr<Plan> table_scan_executor;
        auto conds = std::move(query->where_conds);
        RewriteConds(conds, x->tab_name);

        std::vector<std::string> index_col_names;
        int index_cond = get_index_cols(x->tab_name, conds, index_col_names);

        // separate conditions by whether has subquery
        std::vector<Condition> scan_conds;
        std::vector<Condition> subquery_conds;
        for (auto& cond : conds) {
            if (cond.rhs_type == RhsType::SUB_QUERY) {
                subquery_conds.push_back(cond);
            } else {
                scan_conds.push_back(cond);
            }
        }

        if (index_cond == -1) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executor =
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, index_col_names, conds);
        } else {
            table_scan_executor =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, index_col_names, conds, index_cond);
        }

        std::vector<std::shared_ptr<Plan>> sub_queries;
        for (auto& cond : conds) {
            if (cond.rhs_type == RhsType::SUB_QUERY) {
                sub_queries.emplace_back(generate_select_plan(aggregation_cache, cond.rhs_sub_query, context));
            }
        }
        std::shared_ptr<Plan> filter_executor = !conds.empty()
            ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor), std::move(sub_queries),
                                              conds, x->tab_name, index_cond)
            : std::move(table_scan_executor);

        plannerRoot = std::make_shared<DMLPlan>(T_Delete, filter_executor, x->tab_name, std::vector<Value>(), conds,
                                                std::vector<SetClause>());
        // if (index_cond == -1) {  // 该表没有索引
        //     index_col_names.clear();
        //     table_scan_executor =
        //         std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, index_col_names, scan_conds);
        // } else {
        //     table_scan_executor =
        //         std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, index_col_names, conds,
        //         index_cond);
        // }

        // std::vector<std::shared_ptr<Plan>> sub_queries;
        // for (auto& cond : subquery_conds) {
        //     sub_queries.emplace_back(generate_select_plan(cond.rhs_sub_query, context));
        // }
        // std::shared_ptr<Plan> filter_executor = nullptr;
        // if (table_scan_executor->tag == T_SeqScan) {
        //     filter_executor = !subquery_conds.empty()
        //         ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor),
        //         std::move(sub_queries),
        //                                           subquery_conds, x->tab_name, index_cond)
        //         : std::move(table_scan_executor);

        // } else {
        //     filter_executor = !subquery_conds.empty()
        //         ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor),
        //         std::move(sub_queries),
        //                                           conds, x->tab_name, index_cond)
        //         : std::move(table_scan_executor);
        // }

        // plannerRoot = std::make_shared<DMLPlan>(T_Delete, filter_executor, x->tab_name, std::vector<Value>(), conds,
        //                                         std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(query->parse)) {
        // update;
        // 生成表扫描方式
        std::shared_ptr<Plan> table_scan_executor;
        auto conds = std::move(query->where_conds);
        RewriteConds(conds, x->tab_name);

        std::vector<std::string> index_col_names;
        int index_cond = get_index_cols(x->tab_name, conds, index_col_names);

        // separate conditions by whether has subquery
        std::vector<Condition> scan_conds;
        std::vector<Condition> subquery_conds;
        for (auto& cond : conds) {
            if (cond.rhs_type == RhsType::SUB_QUERY) {
                subquery_conds.push_back(cond);
            } else {
                scan_conds.push_back(cond);
            }
        }

        if (index_cond == -1) {  // 该表没有索引
            index_col_names.clear();
            table_scan_executor =
                std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, index_col_names, conds);
        } else {  // 存在索引
            table_scan_executor =
                std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, index_col_names, conds, index_cond);
        }

        std::vector<std::shared_ptr<Plan>> sub_queries;
        for (auto& cond : conds) {
            if (cond.rhs_type == RhsType::SUB_QUERY) {
                sub_queries.emplace_back(generate_select_plan(aggregation_cache, cond.rhs_sub_query, context));
            }
        }
        std::shared_ptr<Plan> filter_executor = !conds.empty()
            ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor), std::move(sub_queries),
                                              conds, x->tab_name, index_cond)
            : std::move(table_scan_executor);

        plannerRoot = std::make_shared<DMLPlan>(T_Update, filter_executor, x->tab_name, std::vector<Value>(), conds,
                                                query->set_clauses);
        // if (index_cond == -1) {  // 该表没有索引
        //     index_col_names.clear();
        //     table_scan_executor =
        //         std::make_shared<ScanPlan>(T_SeqScan, sm_manager_, x->tab_name, index_col_names, scan_conds);
        // } else {  // 存在索引
        //     table_scan_executor =
        //         std::make_shared<ScanPlan>(T_IndexScan, sm_manager_, x->tab_name, index_col_names, conds,
        //         index_cond);
        // }

        // std::vector<std::shared_ptr<Plan>> sub_queries;
        // for (auto& cond : subquery_conds) {
        //     sub_queries.emplace_back(generate_select_plan(cond.rhs_sub_query, context));
        // }
        // std::shared_ptr<Plan> filter_executor = nullptr;
        // if (table_scan_executor->tag == T_SeqScan) {
        //     filter_executor = !subquery_conds.empty()
        //         ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor),
        //         std::move(sub_queries),
        //                                           subquery_conds, x->tab_name, index_cond)
        //         : std::move(table_scan_executor);

        // } else {
        //     filter_executor = !subquery_conds.empty()
        //         ? std::make_shared<PredicatePlan>(T_Predicate, std::move(table_scan_executor),
        //         std::move(sub_queries),
        //                                           conds, x->tab_name, index_cond)
        //         : std::move(table_scan_executor);
        // }

        // plannerRoot = std::make_shared<DMLPlan>(T_Update, filter_executor, x->tab_name, std::vector<Value>(),
        //                                         query->where_conds, query->set_clauses);
    } else if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(query->parse)) {

        std::shared_ptr<plannerInfo> root = std::make_shared<plannerInfo>(x);
        // 生成select语句的查询执行计划
        std::shared_ptr<Plan> projection = generate_select_plan(aggregation_cache, std::move(query), context);
        plannerRoot = std::make_shared<DMLPlan>(T_select, projection, std::string(), std::vector<Value>(),
                                                std::vector<Condition>(), std::vector<SetClause>());
    } else if (auto x = std::dynamic_pointer_cast<ast::LoadStmt>(query->parse)) {
        // load;
        auto load = std::make_shared<LoadPlan>(T_Load, x->tab_name, x->path);
        plannerRoot = std::make_shared<DMLPlan>(T_Load, load, std::string(), std::vector<Value>(),
                                                std::vector<Condition>(), std::vector<SetClause>());
    } else {
        throw InternalError("Unexpected AST root");
    }
    return plannerRoot;
}