/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "rm_scan.h"
#include "rm_file_handle.h"

RmScan::RmScan(const RmFileHandle* file_handle) : file_handle_(file_handle), rid_() {}

RmScan::~RmScan() {
    auto* page = page_handle_.page;
    if (page != nullptr) {
        file_handle_->buffer_pool_manager_->unpin_page(page->get_page_id(), page->is_dirty());
    }
}

void RmScan::init(std::vector<Condition> conds, std::vector<ColMeta>& cols) {
    conds_ = conds;
    for (size_t i = 0; i < conds_.size(); ++i) {
        auto& cond = conds_[i];

        size_t cmp_times = 1;
        if (cond.rhs_type == RhsType::LIST) {
            cmp_times = cond.rhs_val_list.size();
        }
        value_map_.emplace_back(std::vector<Value>());
        for (size_t j = 0; j < cmp_times; ++j) {
            if (cond.rhs_type == RhsType::VALUE) {
                value_map_[i].push_back(cond.rhs_val);
            } else if (cond.rhs_type == RhsType::COLUMN) {
                value_map_[i].push_back(Value());
            } else if (cond.rhs_type == RhsType::LIST) {
                value_map_[i].push_back(cond.rhs_val_list[j]);
            }
        }

        for (size_t j = 0; j < cols.size(); ++j) {
            if (cols[j].name == cond.lhs_col.col_name) {
                col_map_.emplace_back(cols[j], cols[j].offset);
                break;
            }
        }
    }
}

void RmScan::begin() {
    if (page_handle_.page != nullptr) {
        file_handle_->buffer_pool_manager_->unpin_page(page_handle_.page->get_page_id(), page_handle_.page->is_dirty());
    }

    rid_ = Rid{0, -1};
    int record_size = file_handle_->file_hdr_.bitmap_size * BITMAP_WIDTH;
    bool find = false;
    for (int i = 1, num_pages = file_handle_->file_hdr_.num_pages; i < num_pages; ++i) {
        PageId page_id{file_handle_->fd_, i};
        Page* page = file_handle_->buffer_pool_manager_->fetch_page(page_id);
        page_handle_ = RmPageHandle(&file_handle_->file_hdr_, page);
        int slot_no = Bitmap::first_bit(true, page_handle_.bitmap, record_size);
        while (!find && slot_no != record_size) {
            if (filter(page_handle_.get_slot(slot_no))) {
                find = true;
                break;
            }
            slot_no = Bitmap::next_bit(true, page_handle_.bitmap, record_size, slot_no);
        }
        if (find) {
            rid_ = Rid{i, slot_no};
            break;
        } else {
            file_handle_->buffer_pool_manager_->unpin_page(page_id, false);
        }
    }
}

void RmScan::next() {
    // Todo:
    // 找到文件中下一个存放了记录的非空闲位置，用rid_来指向这个位置
    int record_size = file_handle_->file_hdr_.bitmap_size * BITMAP_WIDTH;
    int slot_no = -1;
    bool find = false;
    for (int i = rid_.page_no, num_pages = file_handle_->file_hdr_.num_pages; i < num_pages; ++i) {
        slot_no = Bitmap::next_bit(true, page_handle_.bitmap, record_size, rid_.slot_no);
        while (!find && slot_no != record_size) {
            if (filter(page_handle_.get_slot(slot_no))) {
                find = true;
                break;
            }
            slot_no = Bitmap::next_bit(true, page_handle_.bitmap, record_size, slot_no);
        }
        if (find) {
            rid_ = Rid{i, slot_no};
            break;
        } else {
            if (i + 1 == num_pages) {
                rid_ = Rid{num_pages, -1};
                break;
            }
            PageId page_id{file_handle_->fd_, i};
            auto* page = page_handle_.page;
            file_handle_->buffer_pool_manager_->unpin_page(page_id, page->is_dirty());
            page_id.page_no = i + 1;
            page = file_handle_->buffer_pool_manager_->fetch_page(page_id);
            page_handle_ = RmPageHandle(&file_handle_->file_hdr_, page);
            rid_ = Rid{i + 1, -1};
        }
    }
}

bool RmScan::is_end() const {
    // Todo: 修改返回值
    if (rid_.page_no == file_handle_->file_hdr_.num_pages || rid_.page_no == 0) {
        return true;
    }
    if (rid_.page_no < file_handle_->file_hdr_.num_pages - 1) {
        return false;
    }
    return !Bitmap::is_set(page_handle_.bitmap, rid_.slot_no);
}

Rid& RmScan::rid() {
    return rid_;
}

bool RmScan::filter(char* record) {
    bool pass = true;
    bool result = false;
    size_t cmp_times = 1;
    for (size_t i = 0; i < conds_.size(); ++i) {
        auto& cond = conds_[i];
        auto& left_col = col_map_[i].first;
        auto left_offset = col_map_[i].second;

        if (cond.op == CompOp::OP_EQ || cond.op == CompOp::OP_NE) {
            if (left_col.type == ColType::TYPE_INT) {
                left_val_.type = ColType::TYPE_INT;
                left_val_.raw_ptr = record + left_offset;
            } else if (left_col.type == ColType::TYPE_FLOAT) {
                left_val_.type = ColType::TYPE_FLOAT;
                left_val_.raw_ptr = record + left_offset;
            } else if (left_col.type == ColType::TYPE_STRING) {
                left_val_.ConvertRecordToValue(record, left_col.type, left_offset, left_col.len);
            }
        } else {
            if (left_col.type == ColType::TYPE_INT) {
                left_val_.type = ColType::TYPE_INT;
                left_val_.int_val = *(int*)(record + left_offset);
            } else if (left_col.type == ColType::TYPE_FLOAT) {
                left_val_.type = ColType::TYPE_FLOAT;
                left_val_.float_val = *(float*)(record + left_offset);
            } else if (left_col.type == ColType::TYPE_STRING) {
                left_val_.ConvertRecordToValue(record, left_col.type, left_offset, left_col.len);
            }
        }

        result = false;
        auto& right_val = value_map_[i][0];
        right_val.raw_ptr = (char*)&right_val.int_val;
        switch (cond.op) {
            case CompOp::OP_EQ: {
                result = left_val_ == right_val;
                break;
            }
            case CompOp::OP_NE: {
                result = left_val_ != right_val;
                break;
            }
            case CompOp::OP_LT: {
                result = left_val_ < right_val;
                break;
            }
            case CompOp::OP_GT: {
                result = left_val_ > right_val;
                break;
            }
            case CompOp::OP_LE: {
                result = left_val_ <= right_val;
                break;
            }
            case CompOp::OP_GE: {
                result = left_val_ >= right_val;
                break;
            }
            case CompOp::OP_IN: {
                result |= left_val_ == right_val;
                break;
            }
            default: {
                throw InternalError("Invalid Condition Compare Operator");
            }
        }

        if (__glibc_unlikely(cond.rhs_type == RhsType::LIST)) {
            cmp_times = cond.rhs_val_list.size();
            for (size_t j = 1; j < cmp_times; ++j) {
                right_val = value_map_[i][j];
                result |= left_val_ == right_val;
            }
        }

        if (!result) {
            pass = false;
            break;
        }
    }
    return pass;
}
