/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else {
        txn->set_state(TransactionState::GROWING);
    }
    LockDataId lock_data_id(tab_fd, rid, LockDataType::RECORD);
    LockRequest lock_request(txn->get_transaction_id(), LockMode::SHARED);
    // 表上是否有IS锁
    bool table_IS_lock = lock_IS_on_table(txn, tab_fd);
    if (table_IS_lock) {
        auto& lock_request_queue = lock_table_[lock_data_id];
        // 当前事务是否已经申请了目标数据项上的锁
        if (lock_table_.find(lock_data_id) != lock_table_.end()) {
            for (auto item : lock_request_queue.request_queue_) {
                if (item.txn_id_ == txn->get_transaction_id()) {
                    if (item.lock_mode_ == LockMode::SHARED || item.lock_mode_ == LockMode::EXLUCSIVE) {
                        return true;
                    }
                }
            }
        }
        // 申请的锁放入到全局锁表中
        txn->get_lock_set()->emplace(lock_data_id);
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK ||
            lock_request_queue.group_lock_mode_ == GroupLockMode::S) {
            lock_request.granted_ = true;
            lock_request_queue.request_queue_.push_back(lock_request);
            lock_request_queue.group_lock_mode_ = GroupLockMode::S;
            lock_request_queue.shared_lock_num_++;
            return true;
        }
        if (conflict(&lock_request_queue, lock_request, txn, LockMode::SHARED)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        // wait-die阻塞
        lock_request_queue.request_queue_.push_back(lock_request);
        lock_request_queue.cv_.notify_all();
        lock_request_queue.cv_.wait(lock, [&]() -> bool {
            if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK ||
                lock_request_queue.group_lock_mode_ == GroupLockMode::S) {
                return true;
            }
            if (txn->get_state() == TransactionState::ABORTED) {
                return true;
            }
            return false;
        });
        if (txn->get_state() == TransactionState::ABORTED) {
            return false;
        }
        if (conflict(&lock_request_queue, lock_request, txn, LockMode::SHARED)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        for (auto item : lock_request_queue.request_queue_) {
            if (item.txn_id_ == txn->get_transaction_id()) {
                item.granted_ = true;
                lock_request_queue.group_lock_mode_ = GroupLockMode::S;
                lock_request_queue.shared_lock_num_++;
                return true;
            }
        }
    }
    // 表上没有IS锁
    else {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }

    return true;
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else {
        txn->set_state(TransactionState::GROWING);
    }
    LockDataId lock_data_id(tab_fd, rid, LockDataType::RECORD);
    LockRequest lock_request(txn->get_transaction_id(), LockMode::EXLUCSIVE);
    // 表上是否有IS锁
    bool table_IX_lock = lock_IS_on_table(txn, tab_fd);
    if (table_IX_lock) {
        auto& lock_request_queue = lock_table_[lock_data_id];
        if (lock_table_.find(lock_data_id) != lock_table_.end()) {
            for (auto item : lock_request_queue.request_queue_) {
                if (item.txn_id_ == txn->get_transaction_id()) {
                    if (item.lock_mode_ == LockMode::SHARED) {
                        lock.unlock();
                        return lock_update_on_record(txn, lock_data_id);
                    } else {
                        return true;
                    }
                }
            }
        }
        txn->get_lock_set()->emplace(lock_data_id);
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
            lock_request.granted_ = true;
            lock_request_queue.group_lock_mode_ = GroupLockMode::X;
            lock_request_queue.shared_lock_num_ = 1;
            return true;
        }
        if (conflict(&lock_request_queue, lock_request, txn, LockMode::EXLUCSIVE)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        lock_request_queue.request_queue_.push_back(lock_request);
        lock_request_queue.cv_.notify_all();
        lock_request_queue.cv_.wait(lock, [&]() -> bool {
            if (txn->get_state() == TransactionState::ABORTED) {
                return true;
            }
            if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
                return true;
            }
            return false;
        });
        if (txn->get_state() == TransactionState::ABORTED) {
            return false;
        }
        if (conflict(&lock_request_queue, lock_request, txn, LockMode::EXLUCSIVE)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        for (auto& item : lock_request_queue.request_queue_) {
            if (item.txn_id_ == txn->get_transaction_id()) {
                item.granted_ = true;
                lock_request_queue.group_lock_mode_ = GroupLockMode::X;
                lock_request_queue.shared_lock_num_ = 1;
                return true;
            }
        }
    } else {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    return true;
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction* txn, int tab_fd) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    LockRequest lock_request(txn->get_transaction_id(), LockMode::SHARED);
    auto& lock_request_queue = lock_table_[lock_data_id];
    if (lock_table_.find(lock_data_id) != lock_table_.end()) {
        for (auto item : lock_request_queue.request_queue_) {
            if (item.txn_id_ == txn->get_transaction_id()) {
                // 升级成SIX锁
                if (item.lock_mode_ == LockMode::INTENTION_EXCLUSIVE) {
                    lock.unlock();
                    return lock_update_SIX_table(txn, lock_data_id);
                }
                // 升级成S锁
                else if (item.lock_mode_ == LockMode::INTENTION_SHARED) {
                    lock.unlock();
                    return lock_update_S_table(txn, lock_data_id);
                } else {
                    return true;
                }
            }
        }
    }
    txn->get_lock_set()->emplace(lock_data_id);
    if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK ||
        lock_request_queue.group_lock_mode_ == GroupLockMode::IS ||
        lock_request_queue.group_lock_mode_ == GroupLockMode::S) {
        lock_request.granted_ = true;
        lock_request_queue.group_lock_mode_ = GroupLockMode::S;
        lock_request_queue.shared_lock_num_++;
        lock_request_queue.request_queue_.push_back(lock_request);
        return true;
    }
    if (conflict(&lock_request_queue, lock_request, txn, LockMode::SHARED)) {
        // txn->set_state(TransactionState::ABORTED);
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_request_queue.request_queue_.push_back(lock_request);
    lock_request_queue.cv_.notify_all();
    lock_request_queue.cv_.wait(lock, [&]() -> bool {
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::IS ||
            lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK ||
            lock_request_queue.group_lock_mode_ == GroupLockMode::S) {
            return lock_request_queue.request_queue_.front().txn_id_ == txn->get_transaction_id();
        }
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (conflict(&lock_request_queue, lock_request, txn, LockMode::SHARED)) {
            // txn->set_state(TransactionState::ABORTED);
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        return lock_request_queue.request_queue_.front().txn_id_ == txn->get_transaction_id();
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }

    for (auto& item : lock_request_queue.request_queue_) {
        if (item.txn_id_ == txn->get_transaction_id()) {
            item.granted_ = true;
            lock_request_queue.group_lock_mode_ = GroupLockMode::S;
            lock_request_queue.shared_lock_num_++;
            return true;
        }
    }

    return true;
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction* txn, int tab_fd) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    LockRequest lock_request(txn->get_transaction_id(), LockMode::EXLUCSIVE);
    auto& lock_request_queue = lock_table_[lock_data_id];
    if (lock_table_.find(lock_data_id) != lock_table_.end()) {
        for (auto item : lock_request_queue.request_queue_) {
            if (item.txn_id_ == txn->get_transaction_id()) {
                if (item.lock_mode_ == LockMode::EXLUCSIVE) {
                    return true;
                }
                // 否则全部升级成X锁
                else {
                    lock.unlock();
                    return lock_update_X_table(txn, lock_data_id);
                }
            }
        }
    }
    txn->get_lock_set()->emplace(lock_data_id);
    if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
        lock_request.granted_ = true;
        lock_request_queue.request_queue_.push_back(lock_request);
        lock_request_queue.group_lock_mode_ = GroupLockMode::X;
        return true;
    }
    if (conflict(&lock_request_queue, lock_request, txn, LockMode::EXLUCSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    // 处理wait—die
    lock_request_queue.request_queue_.push_back(lock_request);
    lock_request_queue.cv_.notify_all();
    lock_request_queue.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
            return lock_request_queue.request_queue_.front().txn_id_ == txn->get_transaction_id();
        }
        if (conflict(&lock_request_queue, lock_request, txn, LockMode::EXLUCSIVE)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        return lock_request_queue.request_queue_.front().txn_id_ == txn->get_transaction_id();
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    for (auto& item : lock_request_queue.request_queue_) {
        if (item.txn_id_ == txn->get_transaction_id()) {
            item.granted_ = true;
            lock_request_queue.group_lock_mode_ = GroupLockMode::X;
            return true;
        }
    }
    return true;
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction* txn, int tab_fd) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else {
        txn->set_state(TransactionState::GROWING);
    }
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    LockRequest lock_request(txn->get_transaction_id(), LockMode::INTENTION_SHARED);
    auto& lock_request_queue = lock_table_[lock_data_id];
    if (lock_table_.find(lock_data_id) != lock_table_.end()) {
        for (auto item : lock_request_queue.request_queue_) {
            // 等级都比IS高
            if (item.txn_id_ == txn->get_transaction_id()) {
                return true;
            }
        }
    }
    // IS锁只和X不共存
    txn->get_lock_set()->emplace(lock_data_id);
    if (lock_request_queue.group_lock_mode_ != GroupLockMode::X) {
        lock_request.granted_ = true;
        lock_request_queue.request_queue_.push_back(lock_request);
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
            lock_request_queue.group_lock_mode_ = GroupLockMode::IS;
        }
        return true;
    }
    if (conflict(&lock_request_queue, lock_request, txn, LockMode::INTENTION_SHARED)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_request_queue.request_queue_.push_back(lock_request);
    lock_request_queue.cv_.notify_all();
    lock_request_queue.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (lock_request_queue.group_lock_mode_ != GroupLockMode::X) {
            return true;
        }
        return false;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (conflict(&lock_request_queue, lock_request, txn, LockMode::INTENTION_SHARED)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    if (lock_request_queue.group_lock_mode_ != GroupLockMode::X) {
        lock_request.granted_ = true;
        // lock_request_queue.request_queue_.push_back(lock_request);
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
            lock_request_queue.group_lock_mode_ = GroupLockMode::IS;
        }
        return true;
    }
    return true;
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction* txn, int tab_fd) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else {
        txn->set_state(TransactionState::GROWING);
    }
    LockDataId lock_data_id(tab_fd, LockDataType::TABLE);
    LockRequest lock_request(txn->get_transaction_id(), LockMode::INTENTION_EXCLUSIVE);
    auto& lock_request_queue = lock_table_[lock_data_id];
    if (lock_table_.find(lock_data_id) != lock_table_.end()) {
        for (auto item : lock_request_queue.request_queue_) {
            if (item.txn_id_ == txn->get_transaction_id()) {
                if (item.lock_mode_ == LockMode::EXLUCSIVE || item.lock_mode_ == LockMode::INTENTION_EXCLUSIVE ||
                    item.lock_mode_ == LockMode::S_IX) {
                    return true;
                }
            }
            if (item.lock_mode_ == LockMode::SHARED) {
                lock.unlock();
                return lock_update_SIX_table(txn, lock_data_id);
            }
            if (item.lock_mode_ == LockMode::INTENTION_SHARED) {
                lock.unlock();
                return lock_update_IX_table(txn, lock_data_id);
            }
        }
    }
    txn->get_lock_set()->emplace(lock_data_id);
    if (lock_request_queue.group_lock_mode_ == GroupLockMode::IS ||
        lock_request_queue.group_lock_mode_ == GroupLockMode::IX ||
        lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
        lock_request.granted_ = true;
        lock_request_queue.group_lock_mode_ = GroupLockMode::IX;
        lock_request_queue.request_queue_.push_back(lock_request);
        lock_request_queue.IX_lock_num_++;
        return true;
    }
    if (conflict(&lock_request_queue, lock_request, txn, LockMode::INTENTION_EXCLUSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_request_queue.request_queue_.push_back(lock_request);
    lock_request_queue.cv_.notify_all();
    lock_request_queue.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (lock_request_queue.group_lock_mode_ == GroupLockMode::IS ||
            lock_request_queue.group_lock_mode_ == GroupLockMode::IX ||
            lock_request_queue.group_lock_mode_ == GroupLockMode::NON_LOCK) {
            return true;
        }
        return false;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (conflict(&lock_request_queue, lock_request, txn, LockMode::INTENTION_EXCLUSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_request.granted_ = true;
    lock_request_queue.group_lock_mode_ = GroupLockMode::IX;
    lock_request_queue.IX_lock_num_++;
    return true;
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction* txn, LockDataId lock_data_id) {
    std::unique_lock<std::mutex> lock(latch_);
    if (txn->get_state() == TransactionState::COMMITTED || txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (txn->get_state() == TransactionState::GROWING) {
        txn->set_state(TransactionState::SHRINKING);
    }
    if (lock_table_.find(lock_data_id) == lock_table_.end()) {
        return true;
    }
    LockRequestQueue& lock_request_queue = lock_table_[lock_data_id];
    for (auto it = lock_request_queue.request_queue_.begin(); it != lock_request_queue.request_queue_.end(); ++it) {
        LockRequest& req = *it;
        if (req.txn_id_ == txn->get_transaction_id()) {
            if (req.lock_mode_ == LockMode::SHARED && req.granted_) {
                lock_request_queue.request_queue_.erase(it);
                if (lock_request_queue.request_queue_.empty()) {
                    lock_table_.erase(lock_data_id);
                } else {
                    lock_request_queue.shared_lock_num_--;
                    update_group_lock_mode(&lock_request_queue);
                    lock_request_queue.cv_.notify_all();
                }
            } else if (req.lock_mode_ == LockMode::INTENTION_EXCLUSIVE && req.granted_) {
                lock_request_queue.request_queue_.erase(it);
                if (lock_request_queue.request_queue_.empty()) {
                    lock_table_.erase(lock_data_id);
                } else {
                    lock_request_queue.IX_lock_num_--;
                    update_group_lock_mode(&lock_request_queue);
                    lock_request_queue.cv_.notify_all();
                }
            } else if (req.lock_mode_ == LockMode::S_IX && req.granted_) {
                lock_request_queue.request_queue_.erase(it);
                if (lock_request_queue.request_queue_.empty()) {
                    lock_table_.erase(lock_data_id);
                } else {
                    lock_request_queue.shared_lock_num_--;
                    lock_request_queue.IX_lock_num_--;
                    update_group_lock_mode(&lock_request_queue);
                    lock_request_queue.cv_.notify_all();
                }
            } else {
                lock_request_queue.request_queue_.erase(it);
                if (lock_request_queue.request_queue_.empty()) {
                    lock_table_.erase(lock_data_id);
                } else {
                    auto before = lock_request_queue.group_lock_mode_;
                    update_group_lock_mode(&lock_request_queue);
                    if (lock_request_queue.group_lock_mode_ != before) {
                        lock_request_queue.cv_.notify_all();
                    }
                }
            }
            return true;
        }
    }
    return true;
}
void LockManager::update_group_lock_mode(LockRequestQueue* lock_request_queue) {
    int c_S = lock_request_queue->shared_lock_num_;
    int c_X = 0;
    int c_IX = lock_request_queue->IX_lock_num_;
    int c_IS = 0;
    int c_SIX = 0;
    for (auto& item : lock_request_queue->request_queue_) {
        if (item.lock_mode_ == LockMode::EXLUCSIVE && item.granted_) {
            c_X++;
        }
        if (item.lock_mode_ == LockMode::S_IX && item.granted_) {
            c_SIX++;
        }
        if (item.lock_mode_ == LockMode::INTENTION_SHARED && item.granted_) {
            c_IS++;
        }
    }
    if (c_X) {
        lock_request_queue->group_lock_mode_ = GroupLockMode::X;
        return;
    } else if (c_SIX) {
        lock_request_queue->group_lock_mode_ = GroupLockMode::SIX;
        return;
    } else if (c_S) {
        lock_request_queue->group_lock_mode_ = GroupLockMode::S;
        return;
    } else if (c_IX) {
        lock_request_queue->group_lock_mode_ = GroupLockMode::IX;
        return;
    } else if (c_IS) {
        lock_request_queue->group_lock_mode_ = GroupLockMode::IS;
        return;
    } else {
        lock_request_queue->group_lock_mode_ = GroupLockMode::NON_LOCK;
        return;
    }
}
bool LockManager::lock_update_on_record(Transaction* txn, LockDataId lock_data_id) {

    std::unique_lock<std::mutex> lock(latch_);
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else {
        txn->set_state(TransactionState::GROWING);
    }
    LockRequestQueue& lock_requests = lock_table_[lock_data_id];
    if (lock_requests.upgrading_) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
        return false;
    }
    if (lock_requests.group_lock_mode_ == GroupLockMode::S && lock_requests.shared_lock_num_ == 1) {
        lock_requests.upgrading_ = true;
        lock_requests.group_lock_mode_ = GroupLockMode::X;
        lock_requests.shared_lock_num_ = 0;
        for (auto& req : lock_requests.request_queue_) {
            if (req.txn_id_ == txn->get_transaction_id()) {
                req.granted_ = true;
                req.lock_mode_ = LockMode::EXLUCSIVE;
                break;
            }
        }
        lock_requests.upgrading_ = false;
        return true;
    }
    if (conflict(&lock_requests, txn, LockMode::EXLUCSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_requests.upgrading_ = true;
    lock_requests.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (lock_requests.group_lock_mode_ == GroupLockMode::S && lock_requests.shared_lock_num_ == 1) {
            return true;
        }
        return false;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (conflict(&lock_requests, txn, LockMode::EXLUCSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    for (auto& req : lock_requests.request_queue_) {
        if (req.txn_id_ == txn->get_transaction_id()) {
            req.granted_ = true;
            req.lock_mode_ = LockMode::EXLUCSIVE;
            lock_requests.group_lock_mode_ = GroupLockMode::X;
            lock_requests.shared_lock_num_ = 0;
            lock_requests.upgrading_ = false;
            return true;
        }
    }
    return true;
}
bool LockManager::lock_update_SIX_table(Transaction* txn, LockDataId lock_data_id) {
    std::unique_lock<std::mutex> lock(latch_);
    if (txn->get_state() == TransactionState::COMMITTED || txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }
    LockRequestQueue& lock_requests = lock_table_[lock_data_id];
    if (lock_requests.upgrading_) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
    }
    if (lock_requests.IX_lock_num_ + lock_requests.shared_lock_num_ == 1) {
        lock_requests.upgrading_ = true;
        for (auto& req : lock_requests.request_queue_) {
            if (req.txn_id_ == txn->get_transaction_id()) {
                req.granted_ = true;
                if (req.lock_mode_ == LockMode::SHARED) {
                    lock_requests.IX_lock_num_++;
                } else {
                    lock_requests.shared_lock_num_++;
                }
                req.lock_mode_ = LockMode::S_IX;
                break;
            }
        }
        lock_requests.group_lock_mode_ = GroupLockMode::SIX;
        lock_requests.upgrading_ = false;
        return true;
    }
    if (conflict(&lock_requests, txn, LockMode::S_IX)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_requests.upgrading_ = true;
    lock_requests.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (lock_requests.IX_lock_num_ + lock_requests.shared_lock_num_ == 1) {
            return true;
        }
        return false;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (conflict(&lock_requests, txn, LockMode::S_IX)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    for (auto& req : lock_requests.request_queue_) {
        if (req.txn_id_ == txn->get_transaction_id()) {
            lock_requests.upgrading_ = true;
            req.granted_ = true;
            if (req.lock_mode_ == LockMode::SHARED) {
                lock_requests.IX_lock_num_++;
            } else {
                lock_requests.shared_lock_num_++;
            }
            req.lock_mode_ = LockMode::S_IX;
            lock_requests.group_lock_mode_ = GroupLockMode::SIX;
            lock_requests.upgrading_ = false;
            return true;
        }
    }
    return true;
}
bool LockManager::lock_update_S_table(Transaction* txn, LockDataId lock_data_id) {
    std::unique_lock<std::mutex> lock(latch_);
    if (txn->get_state() == TransactionState::COMMITTED || txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }
    LockRequestQueue& lock_requests = lock_table_[lock_data_id];
    if (lock_requests.upgrading_) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
    }
    if (lock_requests.group_lock_mode_ == GroupLockMode::S || lock_requests.group_lock_mode_ == GroupLockMode::IS) {
        lock_requests.upgrading_ = true;
        for (auto& req : lock_requests.request_queue_) {
            if (req.txn_id_ == txn->get_transaction_id()) {
                req.granted_ = true;
                req.lock_mode_ = LockMode::SHARED;
                if (req.lock_mode_ == LockMode::INTENTION_EXCLUSIVE) {
                    lock_requests.IX_lock_num_--;
                }
                break;
            }
        }
        lock_requests.group_lock_mode_ = GroupLockMode::S;
        lock_requests.upgrading_ = false;
        return true;
    }
    if (conflict(&lock_requests, txn, LockMode::SHARED)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_requests.upgrading_ = true;
    lock_requests.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (lock_requests.group_lock_mode_ == GroupLockMode::S || lock_requests.group_lock_mode_ == GroupLockMode::IS) {
            return true;
        }
        return false;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (conflict(&lock_requests, txn, LockMode::SHARED)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    for (auto& req : lock_requests.request_queue_) {
        if (req.txn_id_ == txn->get_transaction_id()) {
            req.granted_ = true;
            req.lock_mode_ = LockMode::SHARED;
            if (req.lock_mode_ == LockMode::INTENTION_EXCLUSIVE) {
                lock_requests.IX_lock_num_--;
            }
            lock_requests.group_lock_mode_ = GroupLockMode::S;
            lock_requests.upgrading_ = false;
            return true;
        }
    }
    return true;
}
bool LockManager::lock_update_X_table(Transaction* txn, LockDataId lock_data_id) {
    std::unique_lock<std::mutex> lock(latch_);
    // 将事务设置成增长状态
    if (txn->get_state() == TransactionState::ABORTED || txn->get_state() == TransactionState::COMMITTED) {
        return false;
    } else if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    } else if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }
    auto& lock_request_queue = lock_table_[lock_data_id];
    if (lock_request_queue.upgrading_) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
    }
    // 需要保证只有一个加锁记录才可升级，否则wait
    int c_lock = 0;
    for (auto& item : lock_request_queue.request_queue_) {
        if (item.granted_) {
            c_lock++;
        }
    }
    if (c_lock == 1) {
        lock_request_queue.upgrading_ = true;
        for (auto& item : lock_request_queue.request_queue_) {
            if (item.txn_id_ == txn->get_transaction_id()) {
                item.granted_ = true;
                if (item.lock_mode_ == LockMode::SHARED) {
                    lock_request_queue.shared_lock_num_--;
                } else if (item.lock_mode_ == LockMode::INTENTION_EXCLUSIVE) {
                    lock_request_queue.IX_lock_num_--;
                } else if (item.lock_mode_ == LockMode::S_IX) {
                    lock_request_queue.shared_lock_num_--;
                    lock_request_queue.IX_lock_num_--;
                }
                item.lock_mode_ = LockMode::EXLUCSIVE;
                break;
            }
        }
        lock_request_queue.group_lock_mode_ = GroupLockMode::X;
        lock_request_queue.upgrading_ = false;
        return true;
    }
    if (conflict(&lock_request_queue, txn, LockMode::EXLUCSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_request_queue.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (conflict(&lock_request_queue, txn, LockMode::EXLUCSIVE)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        int c_lock = 0;
        for (auto& item : lock_request_queue.request_queue_) {
            if (item.granted_) {
                c_lock++;
            }
        }
        return c_lock == 1;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    for (auto& item : lock_request_queue.request_queue_) {
        if (item.txn_id_ == txn->get_transaction_id()) {
            item.granted_ = true;
            lock_request_queue.upgrading_ = true;
            if (item.lock_mode_ == LockMode::INTENTION_EXCLUSIVE) {
                lock_request_queue.IX_lock_num_--;
            }
            if (item.lock_mode_ == LockMode::SHARED) {
                lock_request_queue.shared_lock_num_--;
            }
            if (item.lock_mode_ == LockMode::S_IX) {
                lock_request_queue.shared_lock_num_--;
                lock_request_queue.IX_lock_num_--;
            }
            item.lock_mode_ = LockMode::EXLUCSIVE;
            lock_request_queue.group_lock_mode_ = GroupLockMode::X;
            lock_request_queue.upgrading_ = false;
            return true;
        }
    }
    return true;
}
bool LockManager::lock_update_IX_table(Transaction* txn, LockDataId lock_data_id) {
    std::unique_lock<std::mutex> lock(latch_);
    if (txn->get_state() == TransactionState::COMMITTED || txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    if (txn->get_state() == TransactionState::SHRINKING) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
        return false;
    }
    if (txn->get_state() == TransactionState::DEFAULT) {
        txn->set_state(TransactionState::GROWING);
    }
    LockRequestQueue& lock_requests = lock_table_[lock_data_id];
    if (lock_requests.upgrading_) {
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
    }
    if (lock_requests.group_lock_mode_ == GroupLockMode::IS || lock_requests.group_lock_mode_ == GroupLockMode::IX) {
        lock_requests.upgrading_ = true;
        for (auto& req : lock_requests.request_queue_) {
            if (req.txn_id_ == txn->get_transaction_id()) {
                req.granted_ = true;
                req.lock_mode_ = LockMode::INTENTION_EXCLUSIVE;
                break;
            }
        }
        lock_requests.group_lock_mode_ = GroupLockMode::IX;
        lock_requests.IX_lock_num_++;
        lock_requests.upgrading_ = false;
        return true;
    }
    if (conflict(&lock_requests, txn, LockMode::INTENTION_EXCLUSIVE)) {  // no-wait
        throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
    }
    lock_requests.upgrading_ = true;
    lock_requests.cv_.wait(lock, [&]() -> bool {
        if (txn->get_state() == TransactionState::ABORTED) {
            return true;
        }
        if (conflict(&lock_requests, txn, LockMode::INTENTION_EXCLUSIVE)) {  // no-wait
            throw TransactionAbortException(txn->get_transaction_id(), AbortReason::DEADLOCK_PREVENTION);
        }
        if (lock_requests.group_lock_mode_ == GroupLockMode::NON_LOCK ||
            lock_requests.group_lock_mode_ == GroupLockMode::IS ||
            lock_requests.group_lock_mode_ == GroupLockMode::IX) {
            return true;
        }
        return false;
    });
    if (txn->get_state() == TransactionState::ABORTED) {
        return false;
    }
    for (auto& req : lock_requests.request_queue_) {
        if (req.txn_id_ == txn->get_transaction_id()) {
            lock_requests.upgrading_ = true;
            req.granted_ = true;
            req.lock_mode_ = LockMode::INTENTION_EXCLUSIVE;
            lock_requests.group_lock_mode_ = GroupLockMode::IX;
            lock_requests.IX_lock_num_++;
            lock_requests.upgrading_ = false;
            return true;
        }
    }
    return true;
}

bool LockManager::conflict(LockRequestQueue* lock_requests, Transaction* txn, LockMode new_lock_mode) {
    for (auto& req : lock_requests->request_queue_) {
        if (txn->get_transaction_id() > req.txn_id_) {
            if (new_lock_mode == LockMode::INTENTION_SHARED) {
                return req.lock_mode_ == LockMode::EXLUCSIVE;
            }
            if (new_lock_mode == LockMode::INTENTION_EXCLUSIVE) {
                return req.lock_mode_ == LockMode::SHARED || req.lock_mode_ == LockMode::S_IX ||
                    req.lock_mode_ == LockMode::EXLUCSIVE;
            }
            if (new_lock_mode == LockMode::SHARED) {
                return req.lock_mode_ == LockMode::INTENTION_EXCLUSIVE || req.lock_mode_ == LockMode::S_IX ||
                    req.lock_mode_ == LockMode::EXLUCSIVE;
            }
            if (new_lock_mode == LockMode::S_IX) {
                return req.lock_mode_ != LockMode::INTENTION_SHARED;
            }
            if (new_lock_mode == LockMode::EXLUCSIVE) {
                return true;
            }
        }
    }
    return false;
}

bool LockManager::conflict(LockRequestQueue* lock_requests, LockRequest& request, Transaction* txn,
                           LockMode new_lock_mode) {
    for (auto& req : lock_requests->request_queue_) {
        if (req == request) {
            break;
        }
        if (txn->get_transaction_id() > req.txn_id_) {
            if (new_lock_mode == LockMode::INTENTION_SHARED) {
                return req.lock_mode_ == LockMode::EXLUCSIVE;
            }
            if (new_lock_mode == LockMode::INTENTION_EXCLUSIVE) {
                return req.lock_mode_ == LockMode::SHARED || req.lock_mode_ == LockMode::S_IX ||
                    req.lock_mode_ == LockMode::EXLUCSIVE;
            }
            if (new_lock_mode == LockMode::SHARED) {
                return req.lock_mode_ == LockMode::INTENTION_EXCLUSIVE || req.lock_mode_ == LockMode::S_IX ||
                    req.lock_mode_ == LockMode::EXLUCSIVE;
            }
            if (new_lock_mode == LockMode::S_IX) {
                return req.lock_mode_ != LockMode::INTENTION_SHARED;
            }
            if (new_lock_mode == LockMode::EXLUCSIVE) {
                return true;
            }
        }
    }
    return false;
}