/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"
#include "record/rm_file_handle.h"
#include "system/sm_manager.h"

std::unordered_map<txn_id_t, Transaction*> TransactionManager::txn_map = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction* TransactionManager::begin(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 判断传入事务参数是否为空指针
    // 2. 如果为空指针，创建新事务
    // 3. 把开始事务加入到全局事务表中
    // 4. 返回当前事务指针
    std::scoped_lock lock{latch_};
    Transaction* txn_new = nullptr;
    if (txn == nullptr) {
        txn_new = new Transaction(this->get_next_txn_id());
        txn_new->set_state(TransactionState::GROWING);
    } else {
        txn_new = txn;
    }
    txn_id_t txn_id = txn_new->get_transaction_id();
    txn_map[txn_new->get_transaction_id()] = txn_new;
    // TODO: still need to add a begin_log to log_manager(6.17)
    BeginLogRecord begin_log_record(txn_id);
    log_manager->add_log_to_buffer(&begin_log_record);
    return txn_new;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 如果存在未提交的写操作，提交所有的写操作
    // 2. 释放所有锁
    // 3. 释放事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态
    std::scoped_lock lock{latch_};
    if (txn == nullptr)
        return;

    txn->set_state(TransactionState::SHRINKING);

    std::shared_ptr<std::unordered_set<LockDataId>> lock_set = txn->get_lock_set();
    auto write_set = txn->get_write_set();
    while (!write_set->empty()) {
        auto item = write_set->back();
        write_set->pop_back();
        delete item;
    }
    write_set->clear();

    // 释放锁集
    for (auto lock = lock_set->begin(); lock != lock_set->end(); lock++) {
        this->lock_manager_->unlock(txn, *lock);
    }

    lock_set->clear();

    // TODO: till need to add a commit_log(6.17)
    CommitLogRecord commit_log_record(txn->get_transaction_id());
    commit_log_record.prev_lsn_ = txn->get_prev_lsn();
    log_manager->add_log_to_buffer(&commit_log_record);
    log_manager->flush_log_to_disk();

    txn->set_state(TransactionState::COMMITTED);

    txn_map.erase(txn->get_transaction_id());
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction* txn, LogManager* log_manager) {
    // Todo:
    // 1. 回滚所有写操作
    // 2. 释放所有锁
    // 3. 清空事务相关资源，eg.锁集
    // 4. 把事务日志刷入磁盘中
    // 5. 更新事务状态

    // TODO: still need to rollback(6.17)
    std::scoped_lock lock{latch_};
    if (txn == nullptr)
        return;
    txn->set_state(TransactionState::SHRINKING);
    // 回滚所有写操作
    auto write_set = txn->get_write_set();
    while (!write_set->empty()) {
        auto wr = write_set->back();
        auto tab_name = wr->GetTableName();  // 表名字

        auto rid = wr->GetRid();                                   // 记录标识符
        TabMeta& tab_meta = sm_manager_->db_.get_table(tab_name);  // 表的元数据信息

        // ？？需要考虑是否必要
        auto txn_context = new Context(lock_manager_, log_manager, txn);
        auto write_type = wr->GetWriteType();
        // 回滚的操作为Insert时，此时需要删除相应的记录
        if (write_type == WType::INSERT_TUPLE) {
            auto record = sm_manager_->fhs_.at(tab_name).get()->get_record(rid, txn_context);
            // 删除索引
            for (size_t i = 0; i < tab_meta.indexes.size(); ++i) {
                auto index = tab_meta.indexes[i];
                // 获取索引句柄
                auto ih =
                    sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name, index.cols)).get();
                // 构建索引键
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for (size_t i = 0; i < index.col_num; ++i) {
                    memcpy(key + offset, record->data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->delete_entry(key, txn);
                // 释放资源
                delete[] key;
            }
            sm_manager_->fhs_.at(tab_name).get()->delete_record(rid, txn_context);
            DeleteLogRecord delete_log_record(txn->get_transaction_id(), txn->get_prev_lsn(), *record.get(), rid,
                                              tab_name);
            auto lsn = log_manager->add_log_to_buffer(&delete_log_record);
            txn->set_prev_lsn(lsn);
        } else if (write_type == WType::DELETE_TUPLE) {

            auto record = wr->GetRecord();
            // 删除索引
            for (size_t i = 0; i < tab_meta.indexes.size(); i++) {
                auto index = tab_meta.indexes[i];
                // 获取索引句柄
                auto ih =
                    sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name, index.cols)).get();
                // 构建索引键
                char* key = new char[index.col_tot_len];
                int offset = 0;
                for (size_t i = 0; i < index.col_num; i++) {
                    memcpy(key + offset, record.data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->insert_entry(key, rid, txn);
                // 释放资源
                delete[] key;
            }

            sm_manager_->fhs_.at(tab_name).get()->insert_record(rid, record.data);
            InsertLogRecord insert_log_record(txn->get_transaction_id(), txn->get_prev_lsn(), record, rid, tab_name);
            auto lsn = log_manager->add_log_to_buffer(&insert_log_record);
            txn->set_prev_lsn(lsn);
        } else if (write_type == WType::UPDATE_TUPLE) {
            auto old_value = wr->GetRecord();
            auto record = sm_manager_->fhs_.at(tab_name)->get_record(rid, txn_context);
            for (size_t i = 0; i < tab_meta.indexes.size(); i++) {
                auto index = tab_meta.indexes[i];
                // 获取索引句柄
                auto ih =
                    sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name, index.cols)).get();
                // 构建索引键
                char* del_key = new char[index.col_tot_len];
                char* add_key = new char[index.col_tot_len];
                int offset = 0;
                for (size_t i = 0; i < index.col_num; i++) {
                    memcpy(del_key + offset, record->data + index.cols[i].offset, index.cols[i].len);
                    memcpy(add_key + offset, old_value.data + index.cols[i].offset, index.cols[i].len);
                    offset += index.cols[i].len;
                }
                ih->delete_entry(del_key, txn);
                ih->insert_entry(add_key, rid, txn);
                // 释放资源
                delete[] del_key;
                delete[] add_key;
            }
            sm_manager_->fhs_.at(tab_name).get()->update_record(rid, old_value.data, txn_context);
            UpdateLogRecord update_log_record(txn->get_transaction_id(), txn->get_prev_lsn(), *record.get(), old_value,
                                              rid, tab_name);
            auto lsn = log_manager->add_log_to_buffer(&update_log_record);
            txn->set_prev_lsn(lsn);
        }
        write_set->pop_back();
        delete wr;
    }
    std::shared_ptr<std::unordered_set<LockDataId>> lock_set = txn->get_lock_set();
    // 释放锁集
    for (auto lock = lock_set->begin(); lock != lock_set->end(); lock++) {
        this->lock_manager_->unlock(txn, *lock);
    }
    lock_set->clear();
    // 事务日志的刷入磁盘
    AbortLogRecord abort_log_record(txn->get_transaction_id(), txn->get_prev_lsn());
    abort_log_record.prev_lsn_ = txn->get_prev_lsn();
    log_manager->add_log_to_buffer(&abort_log_record);
    log_manager->flush_log_to_disk();
    txn->set_state(TransactionState::ABORTED);

    txn_map.erase(txn->get_transaction_id());
}